import pytest

import kopf
from kopf.reactor.causation import CREATE, UPDATE, DELETE
from kopf.reactor.handling import subregistry_var
from kopf.reactor.registries import Resource, SimpleRegistry, GlobalRegistry


def test_on_create_minimal(mocker):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=CREATE)

    @kopf.on.create('group', 'version', 'plural')
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == CREATE
    assert handlers[0].field is None
    assert handlers[0].timeout is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None


def test_on_update_minimal(mocker):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=UPDATE)

    @kopf.on.update('group', 'version', 'plural')
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == UPDATE
    assert handlers[0].field is None
    assert handlers[0].timeout is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None


def test_on_delete_minimal(mocker):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=DELETE)

    @kopf.on.delete('group', 'version', 'plural')
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == DELETE
    assert handlers[0].field is None
    assert handlers[0].timeout is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None


def test_on_field_minimal(mocker):
    registry = kopf.get_default_registry()
    resource = Resource('group', 'version', 'plural')
    diff = [('op', ('field', 'subfield'), 'old', 'new')]
    cause = mocker.MagicMock(resource=resource, event=UPDATE, diff=diff)

    @kopf.on.field('group', 'version', 'plural', 'field.subfield')
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event is None
    assert handlers[0].field == ('field', 'subfield')
    assert handlers[0].timeout is None
    assert handlers[0].labels is None
    assert handlers[0].annotations is None


def test_on_field_fails_without_field():
    with pytest.raises(TypeError):
        @kopf.on.field('group', 'version', 'plural')
        def fn(**_):
            pass


def test_on_create_with_all_kwargs(mocker):
    registry = GlobalRegistry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=CREATE)
    mocker.patch('kopf.structs.filters.match', return_value=True)

    @kopf.on.create('group', 'version', 'plural',
                    id='id', timeout=123, registry=registry,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'})
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == CREATE
    assert handlers[0].field is None
    assert handlers[0].id == 'id'
    assert handlers[0].timeout == 123
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}


def test_on_update_with_all_kwargs(mocker):
    registry = GlobalRegistry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=UPDATE)
    mocker.patch('kopf.structs.filters.match', return_value=True)

    @kopf.on.update('group', 'version', 'plural',
                    id='id', timeout=123, registry=registry,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'})
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == UPDATE
    assert handlers[0].field is None
    assert handlers[0].id == 'id'
    assert handlers[0].timeout == 123
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}


@pytest.mark.parametrize('optional', [
    pytest.param(True, id='optional'),
    pytest.param(False, id='mandatory'),
])
def test_on_delete_with_all_kwargs(mocker, optional):
    registry = GlobalRegistry()
    resource = Resource('group', 'version', 'plural')
    cause = mocker.MagicMock(resource=resource, event=DELETE)
    mocker.patch('kopf.structs.filters.match', return_value=True)

    @kopf.on.delete('group', 'version', 'plural',
                    id='id', timeout=123, registry=registry, optional=optional,
                    labels={'somelabel': 'somevalue'},
                    annotations={'someanno': 'somevalue'})
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event == DELETE
    assert handlers[0].field is None
    assert handlers[0].id == 'id'
    assert handlers[0].timeout == 123
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}


def test_on_field_with_all_kwargs(mocker):
    registry = GlobalRegistry()
    resource = Resource('group', 'version', 'plural')
    diff = [('op', ('field', 'subfield'), 'old', 'new')]
    cause = mocker.MagicMock(resource=resource, event=UPDATE, diff=diff)
    mocker.patch('kopf.structs.filters.match', return_value=True)

    @kopf.on.field('group', 'version', 'plural', 'field.subfield',
                   id='id', timeout=123, registry=registry,
                   labels={'somelabel': 'somevalue'},
                   annotations={'someanno': 'somevalue'})
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn
    assert handlers[0].event is None
    assert handlers[0].field ==('field', 'subfield')
    assert handlers[0].id == 'id/field.subfield'
    assert handlers[0].timeout == 123
    assert handlers[0].labels == {'somelabel': 'somevalue'}
    assert handlers[0].annotations == {'someanno': 'somevalue'}


def test_subhandler_declaratively(mocker):
    cause = mocker.MagicMock(event=UPDATE, diff=None)

    registry = SimpleRegistry()
    subregistry_var.set(registry)

    @kopf.on.this()
    def fn(**_):
        pass

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn


def test_subhandler_imperatively(mocker):
    cause = mocker.MagicMock(event=UPDATE, diff=None)

    registry = SimpleRegistry()
    subregistry_var.set(registry)

    def fn(**_):
        pass
    kopf.register(fn)

    handlers = registry.get_cause_handlers(cause)
    assert len(handlers) == 1
    assert handlers[0].fn is fn



from typing import Type

import pykube

from kopf.clients import auth


def _make_cls(resource) -> Type[pykube.objects.APIObject]:
    api = auth.get_pykube_api()
    api_resources = api.resource_list(resource.api_version)['resources']
    resource_kind = next((r['kind'] for r in api_resources if r['name'] == resource.plural), None)
    is_namespaced = next((r['namespaced'] for r in api_resources if r['name'] == resource.plural), None)
    if not resource_kind:
        raise pykube.ObjectDoesNotExist(f"No such CRD: {resource.name}")

    cls_name = resource.plural
    cls_base = pykube.objects.NamespacedAPIObject if is_namespaced else pykube.objects.APIObject
    cls = type(cls_name, (cls_base,), {
        'version': resource.api_version,
        'endpoint': resource.plural,
        'kind': resource_kind,
    })
    return cls



import asyncio

import pytest

import kopf
from kopf.reactor.causation import HANDLER_CAUSES, CREATE, UPDATE, DELETE, RESUME
from kopf.reactor.handling import custom_object_handler


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_1st_step_stores_progress_by_patching(
        registry, handlers, extrahandlers,
        resource, cause_mock, cause_type, k8s_mocked):
    name1 = f'{cause_type}_fn'
    name2 = f'{cause_type}_fn2'

    cause_mock.event = cause_type

    await custom_object_handler(
        lifecycle=kopf.lifecycles.asap,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )

    assert handlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert handlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert handlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert handlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is not None

    assert patch['status']['kopf']['progress'][name1]['retries'] == 1
    assert patch['status']['kopf']['progress'][name1]['success']  # evals to true

    assert 'retries' not in patch['status']['kopf']['progress'][name2]
    assert 'success' not in patch['status']['kopf']['progress'][name2]

    assert 'started' in patch['status']['kopf']['progress'][name1]
    assert 'started' in patch['status']['kopf']['progress'][name2]


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_2nd_step_finishes_the_handlers(
        registry, handlers, extrahandlers,
        resource, cause_mock, cause_type, k8s_mocked):
    name1 = f'{cause_type}_fn'
    name2 = f'{cause_type}_fn2'

    cause_mock.event = cause_type
    cause_mock.body.update({
        'status': {'kopf': {'progress': {
            name1: {'started': '1979-01-01T00:00:00', 'success': True},
            name2: {'started': '1979-01-01T00:00:00'},
        }}}
    })

    await custom_object_handler(
        lifecycle=kopf.lifecycles.one_by_one,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )

    assert extrahandlers.create_mock.call_count == (1 if cause_type == CREATE else 0)
    assert extrahandlers.update_mock.call_count == (1 if cause_type == UPDATE else 0)
    assert extrahandlers.delete_mock.call_count == (1 if cause_type == DELETE else 0)
    assert extrahandlers.resume_mock.call_count == (1 if cause_type == RESUME else 0)

    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert patch['status']['kopf']['progress'] is None



"""
The routines to manipulate the handler progression over the event cycle.

Used to track which handlers are finished, which are not yet,
and how many retries were there.

There could be more than one low-level k8s watch-events per one actual
high-level kopf-event (a cause). The handlers are called at different times,
and the overall handling routine should persist the handler status somewhere.

The structure is this:

.. code-block: yaml

    metainfo: ...
    spec: ...
    status: ...
        kopf:
            progress:
                handler1:
                    started: 2018-12-31T23:59:59,999999
                    stopped: 2018-01-01T12:34:56,789000
                    success: abcdef1234567890fedcba
                handler2:
                    started: 2018-12-31T23:59:59,999999
                    stopped: 2018-01-01T12:34:56,789000
                    failure: abcdef1234567890fedcba
                    message: "Error message."
                handler3:
                    started: 2018-12-31T23:59:59,999999
                    retries: 30
                handler3/sub1:
                    started: 2018-12-31T23:59:59,999999
                    delayed: 2018-01-01T12:34:56,789000
                    retries: 10
                    message: "Not ready yet."
                handler3/sub2:
                    started: 2018-12-31T23:59:59,999999

* ``status.kopf.progress`` stores the state of each individual handler in the
  current handling cycle (by their ids, which are usually the function names).

For each handler's status, the following is stored:

* ``started``: when the handler was attempted for the first time (used for timeouts & durations).
* ``stopped``: when the handler has failed or succeeded.
* ``delayed``: when the handler can retry again.
* ``retries``: the number of retried attempted so far (including reruns and successful attempts).
* ``success``: a digest of where the handler has succeeded (and thus no retries are needed).
* ``failure``: a digest of where the handler has failed completely (no retries will be done).
* ``message``: a brief error message from the last exception (as a hint).

When the full event cycle is executed (possibly including multiple re-runs),
the whole ``status.kopf`` section is purged. The life-long persistence of status
is not intended: otherwise, multiple distinct causes will clutter the status
and collide with each other (especially critical for multiple updates).

The digest of each handler's success or failure can be considered a "version"
of an object being handled, as it was when the handler has finished.
If the object is changed during the remaining handling cycle, the digest
of the finished handlers will be mismatching the actual digest of the object,
and so they will be re-executed.

This is conceptually close to *reconciliation*: the handling is finished
only when all handlers are executed on the latest state of the object.

Creation is treated specially: the creation handlers will never be re-executed.
In case of changes during the creation handling, the remaining creation handlers
will get the new state (as normally), and then there will be an update cycle
with all the changes since the first creation handler -- i.e. not from the last
handler as usually, when the last-seen state is updated.

Update handlers are assumed to be idempotent by concept, so it should be safe
to call them with the changes that are already reflected in the system by some
of the creation handlers.

Note: The Kubernetes-provided "resource version" of the object is not used,
as it increases with every change of the object, while this digest is used
only for the changes relevant to the operator and framework (see `get_state`).
"""

import collections.abc
import copy
import datetime


def is_started(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    return handler.id in progress


def is_sleeping(*, body, digest, handler):
    ts = get_awake_time(body=body, handler=handler)
    finished = is_finished(body=body, digest=digest, handler=handler)
    return not finished and ts is not None and ts > datetime.datetime.utcnow()


def is_awakened(*, body, digest, handler):
    finished = is_finished(body=body, digest=digest, handler=handler)
    sleeping = is_sleeping(body=body, digest=digest, handler=handler)
    return bool(not finished and not sleeping)


def is_finished(*, body, digest, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    success = progress.get(handler.id, {}).get('success', None)
    failure = progress.get(handler.id, {}).get('failure', None)
    return ((success is not None and (success is True or success == digest)) or
            (failure is not None and (failure is True or failure == digest)))


def get_start_time(*, body, patch, handler):
    progress = patch.get('status', {}).get('kopf', {}).get('progress', {})
    new_value = progress.get(handler.id, {}).get('started', None)
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    old_value = progress.get(handler.id, {}).get('started', None)
    value = new_value or old_value
    return None if value is None else datetime.datetime.fromisoformat(value)


def get_awake_time(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    value = progress.get(handler.id, {}).get('delayed', None)
    return None if value is None else datetime.datetime.fromisoformat(value)


def get_retry_count(*, body, handler):
    progress = body.get('status', {}).get('kopf', {}).get('progress', {})
    return progress.get(handler.id, {}).get('retries', None) or 0


def set_start_time(*, body, patch, handler):
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'started': datetime.datetime.utcnow().isoformat(),
    })


def set_awake_time(*, body, patch, handler, delay=None):
    if delay is not None:
        ts = datetime.datetime.utcnow() + datetime.timedelta(seconds=delay)
        ts = ts.isoformat()
    else:
        ts = None
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'delayed': ts,
    })


def set_retry_time(*, body, patch, handler, delay=None):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'retries': retry + 1,
    })
    set_awake_time(body=body, patch=patch, handler=handler, delay=delay)


def store_failure(*, body, patch, digest, handler, exc):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'stopped': datetime.datetime.utcnow().isoformat(),
        'failure': digest,
        'retries': retry + 1,
        'message': f'{exc}',
    })


def store_success(*, body, patch, digest, handler, result=None):
    retry = get_retry_count(body=body, handler=handler)
    progress = patch.setdefault('status', {}).setdefault('kopf', {}).setdefault('progress', {})
    progress.setdefault(handler.id, {}).update({
        'stopped': datetime.datetime.utcnow().isoformat(),
        'success': digest,
        'retries': retry + 1,
        'message': None,
    })
    store_result(patch=patch, handler=handler, result=result)


def store_result(*, patch, handler, result=None):
    if result is None:
        pass
    elif isinstance(result, collections.abc.Mapping):
        # TODO: merge recursively (patch-merge), do not overwrite the keys if they are present.
        patch.setdefault('status', {}).setdefault(handler.id, {}).update(result)
    else:
        # TODO? Fail if already present?
        patch.setdefault('status', {})[handler.id] = copy.deepcopy(result)


def purge_progress(*, body, patch):
    patch.setdefault('status', {}).setdefault('kopf', {})['progress'] = None



import asyncio
import dataclasses
import io
import json
import logging
import os
import re
import sys
import time
from unittest.mock import Mock

import asynctest
import pykube
import pytest
import pytest_mock

from kopf.config import configure
from kopf.engines.logging import ObjectPrefixingFormatter
from kopf.reactor.registries import Resource


def pytest_configure(config):
    config.addinivalue_line('markers', "e2e: end-to-end tests with real operators.")
    config.addinivalue_line('markers', "resource_clustered: (internal parameterizatiom mark).")


def pytest_addoption(parser):
    parser.addoption("--only-e2e", action="store_true", help="Execute end-to-end tests only.")
    parser.addoption("--with-e2e", action="store_true", help="Include end-to-end tests.")


# This logic is not applied if pytest is started explicitly on ./examples/.
# In that case, regular pytest behaviour applies -- this is intended.
def pytest_collection_modifyitems(config, items):

    # Make all tests in this directory and below asyncio-compatible by default.
    for item in items:
        if asyncio.iscoroutinefunction(item.function):
            item.add_marker('asyncio')

    # Put all e2e tests to the end, as they are assumed to be slow.
    def _is_e2e(item):
        path = item.location[0]
        return path.startswith('tests/e2e/') or path.startswith('examples/')
    etc = [item for item in items if not _is_e2e(item)]
    e2e = [item for item in items if _is_e2e(item)]

    # Mark all e2e tests, no matter how they were detected. Just for filtering.
    mark_e2e = pytest.mark.e2e
    for item in e2e:
        item.add_marker(mark_e2e)

    # Minikube tests are heavy and require a cluster. Skip them by default,
    # so that the contributors can run pytest without initial tweaks.
    mark_skip = pytest.mark.skip(reason="E2E tests are not enabled. "
                                        "Use --with-e2e/--only-e2e to enable.")
    if not config.getoption('--with-e2e') and not config.getoption('--only-e2e'):
        for item in e2e:
            item.add_marker(mark_skip)

    # Minify the test-plan if only e2e are requested (all other should be skipped).
    if config.getoption('--only-e2e'):
        items[:] = e2e
    else:
        items[:] = etc + e2e


# Substitute the regular mock with the async-aware mock in the `mocker` fixture.
@pytest.fixture(scope='session', autouse=True)
def enforce_asyncio_mocker():
    pytest_mock._get_mock_module = lambda config: asynctest


@pytest.fixture()
def resource():
    """ The resource used in the tests. Usually mocked, so it does not matter. """
    return Resource('zalando.org', 'v1', 'kopfexamples')

#
# Mocks for Kubernetes API clients (any of them). Reasons:
# 1. We do not test the clients, we test the layers on top of them,
#    so everything low-level should be mocked and assumed to be functional.
# 2. No external calls must be made under any circumstances.
#    The unit-tests must be fully isolated from the environment.
#

@pytest.fixture()
def req_mock(mocker, resource, request):

    # Pykube config is needed to create a pykube's API instance.
    # But we do not want and do not need to actually authenticate, so we mock.
    # Some fields are used by pykube's objects: we have to know them ("leaky abstractions").
    cfg_mock = mocker.patch('kopf.clients.auth.get_pykube_cfg').return_value
    cfg_mock.cluster = {'server': 'localhost'}
    cfg_mock.namespace = 'default'

    # Simulated list of cluster-defined CRDs: all of them at once. See: `resource` fixture(s).
    # Simulate the resource as cluster-scoped is there is a marker on the test.
    namespaced = not any(marker.name == 'resource_clustered' for marker in request.node.own_markers)
    res_mock = mocker.patch('pykube.http.HTTPClient.resource_list')
    res_mock.return_value = {'resources': [
        {'name': 'kopfexamples', 'kind': 'KopfExample', 'namespaced': namespaced},
    ]}

    # Prevent ANY outer requests, no matter what. These ones are usually asserted.
    req_mock = mocker.patch('requests.Session').return_value
    return req_mock


@pytest.fixture()
def stream(req_mock):
    """ A mock for the stream of events as if returned by K8s client. """
    def feed(*args):
        side_effect = []
        for arg in args:
            if isinstance(arg, (list, tuple)):
                arg = iter(json.dumps(event).encode('utf-8') for event in arg)
            side_effect.append(arg)
        req_mock.get.return_value.iter_lines.side_effect = side_effect
    return Mock(spec_set=['feed'], feed=feed)

#
# Mocks for login & checks. Used in specifialised login tests,
# and in all CLI tests (since login is implicit with CLI commands).
#

@dataclasses.dataclass(frozen=True, eq=False, order=False)
class LoginMocks:
    pykube_in_cluster: Mock = None
    pykube_from_file: Mock = None
    pykube_checker: Mock = None
    client_in_cluster: Mock = None
    client_from_file: Mock = None
    client_checker: Mock = None


@pytest.fixture()
def login_mocks(mocker):

    # Pykube config is needed to create a pykube's API instance.
    # But we do not want and do not need to actually authenticate, so we mock.
    # Some fields are used by pykube's objects: we have to know them ("leaky abstractions").
    cfg_mock = mocker.patch('kopf.clients.auth.get_pykube_cfg').return_value
    cfg_mock.cluster = {'server': 'localhost'}
    cfg_mock.namespace = 'default'

    # Make all client libraries potentially optional, but do not skip the tests:
    # skipping the tests is the tests' decision, not this mocking fixture's one.
    kwargs = {}
    try:
        import pykube
    except ImportError:
        pass
    else:
        kwargs.update(
            pykube_in_cluster=mocker.patch.object(pykube.KubeConfig, 'from_service_account'),
            pykube_from_file=mocker.patch.object(pykube.KubeConfig, 'from_file'),
            pykube_checker=mocker.patch.object(pykube.http.HTTPClient, 'get'),
        )
    try:
        import kubernetes
    except ImportError:
        pass
    else:
        kwargs.update(
            client_in_cluster=mocker.patch.object(kubernetes.config, 'load_incluster_config'),
            client_from_file=mocker.patch.object(kubernetes.config, 'load_kube_config'),
            client_checker=mocker.patch.object(kubernetes.client, 'CoreApi'),
        )
    return LoginMocks(**kwargs)

#
# Simulating that Kubernetes client library is not installed.
#

class ProhibitedImportFinder:
    def find_spec(self, fullname, path, target=None):
        if fullname == 'kubernetes' or fullname.startswith('kubernetes.'):
            raise ImportError("Import is prohibited for tests.")


@pytest.fixture()
def _kubernetes():
    # If kubernetes client is required, it should either be installed,
    # or skip the test: we cannot simulate its presence (unlike its absence).
    return pytest.importorskip('kubernetes')


@pytest.fixture()
def _no_kubernetes():
    try:
        import kubernetes as kubernetes_before
    except ImportError:
        yield
        return  # nothing to patch & restore.

    # Remove any cached modules.
    preserved = {}
    for name, mod in list(sys.modules.items()):
        if name == 'kubernetes' or name.startswith('kubernetes.'):
            preserved[name] = mod
            del sys.modules[name]

    # Inject the prohibition for loading this module. And restore when done.
    finder = ProhibitedImportFinder()
    sys.meta_path.insert(0, finder)
    try:
        yield
    finally:
        sys.meta_path.remove(finder)
        sys.modules.update(preserved)

        # Verify if it works and that we didn't break the importing machinery.
        import kubernetes as kubernetes_after
        assert kubernetes_after is kubernetes_before


@pytest.fixture(params=[True], ids=['with-client'])  # for hinting suffixes
def kubernetes(request):
    return request.getfixturevalue('_kubernetes')


@pytest.fixture(params=[False], ids=['no-client'])  # for hinting suffixes
def no_kubernetes(request):
    return request.getfixturevalue('_no_kubernetes')


@pytest.fixture(params=[False, True], ids=['no-client', 'with-client'])
def any_kubernetes(request):
    if request.param:
        return request.getfixturevalue('_kubernetes')
    else:
        return request.getfixturevalue('_no_kubernetes')

#
# Helpers for the timing checks.
#

@pytest.fixture()
def timer():
    return Timer()


class Timer(object):
    """
    A helper context manager to measure the time of the code-blocks.
    Also, supports direct comparison with time-deltas and the numbers of seconds.

    Usage:

        with Timer() as timer:
            do_something()
            print(f"Executing for {timer.seconds}s already.")
            do_something_else()

        print(f"Executed in {timer.seconds}s.")
        assert timer < 5.0
    """

    def __init__(self):
        super().__init__()
        self._ts = None
        self._te = None

    @property
    def seconds(self):
        if self._ts is None:
            return None
        elif self._te is None:
            return time.perf_counter() - self._ts
        else:
            return self._te - self._ts

    def __repr__(self):
        status = 'new' if self._ts is None else 'running' if self._te is None else 'finished'
        return f'<Timer: {self.seconds}s ({status})>'

    def __enter__(self):
        self._ts = time.perf_counter()
        self._te = None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._te = time.perf_counter()

    def __int__(self):
        return int(self.seconds)

    def __float__(self):
        return float(self.seconds)

#
# Helpers for the logging checks.
#


@pytest.fixture()
def logstream(caplog):
    """ Prefixing is done at the final output. We have to intercept it. """

    logger = logging.getLogger()
    handlers = list(logger.handlers)

    configure(verbose=True)

    stream = io.StringIO()
    handler = logging.StreamHandler(stream)
    formatter = ObjectPrefixingFormatter('prefix %(message)s')
    handler.setFormatter(formatter)

    logger.addHandler(handler)
    try:
        with caplog.at_level(logging.DEBUG):
            yield stream
    finally:
        logger.removeHandler(handler)
        logger.handlers[:] = handlers  # undo `configure()`


@pytest.fixture()
def assert_logs(caplog):
    """
    A function to assert the logs are present (by pattern).

    The listed message patterns MUST be present, in the order specified.
    Some other log messages can also be present, but they are ignored.
    """
    def assert_logs_fn(patterns, prohibited=[], strict=False):
        __traceback_hide__ = True
        remaining_patterns = list(patterns)
        for message in caplog.messages:
            # The expected pattern is at position 0.
            # Looking-ahead: if one of the following patterns matches, while the
            # 0th does not, then the log message is missing, and we fail the test.
            for idx, pattern in enumerate(remaining_patterns):
                m = re.search(pattern, message)
                if m:
                    if idx == 0:
                        remaining_patterns[:1] = []
                        break  # out of `remaining_patterns` cycle
                    else:
                        skipped_patterns = remaining_patterns[:idx]
                        raise AssertionError(f"Few patterns were skipped: {skipped_patterns!r}")
                elif strict:
                    raise AssertionError(f"Unexpected log message: {message!r}")

            # Check that the prohibited patterns do not appear in any message.
            for pattern in prohibited:
                m = re.search(pattern, message)
                if m:
                    raise AssertionError(f"Prohibited log pattern found: {message!r} ~ {pattern!r}")

        # If all patterns have been matched in order, we are done.
        # if some are left, but the messages are over, then we fail.
        if remaining_patterns:
            raise AssertionError(f"Few patterns were missed: {remaining_patterns!r}")

    return assert_logs_fn


#
# Helpers for asyncio checks.
#
@pytest.fixture(autouse=True)
def _no_asyncio_pending_tasks():
    """
    Ensure there are no unattended asyncio tasks after the test.

    It looks  both in the test's main event-loop, and in all other event-loops,
    such as the background thread of `KopfRunner` (used in e2e tests).

    Current solution uses some internals of asyncio, since there is no public
    interface for that. The warnings are printed only at the end of pytest.

    An alternative way: set event-loop's exception handler, force garbage
    collection after every test, and check messages from `asyncio.Task.__del__`.
    This, however, requires intercepting all event-loop creation in the code.
    """
    # See `asyncio.all_tasks()` implementation for reference.
    before = {t for t in list(asyncio.tasks._all_tasks) if not t.done()}
    yield
    after = {t for t in list(asyncio.tasks._all_tasks) if not t.done()}
    remains = after - before
    if remains:
        pytest.fail(f"Unattended asyncio tasks detected: {remains!r}")



from kopf import BaseRegistry, SimpleRegistry, GlobalRegistry


def test_creation_of_simple_no_prefix():
    registry = SimpleRegistry()
    assert isinstance(registry, BaseRegistry)
    assert isinstance(registry, SimpleRegistry)
    assert registry.prefix is None


def test_creation_of_simple_with_prefix_argument():
    registry = SimpleRegistry('hello')
    assert registry.prefix == 'hello'


def test_creation_of_simple_with_prefix_keyword():
    registry = SimpleRegistry(prefix='hello')
    assert registry.prefix == 'hello'


def test_creation_of_global():
    registry = GlobalRegistry()
    assert isinstance(registry, BaseRegistry)
    assert isinstance(registry, GlobalRegistry)



import collections

from kopf.reactor.registries import Resource, GlobalRegistry


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


def test_resources():
    registry = GlobalRegistry()
    registry.register_cause_handler('group1', 'version1', 'plural1', some_fn)
    registry.register_cause_handler('group2', 'version2', 'plural2', some_fn)

    resources = registry.resources

    assert isinstance(resources, collections.abc.Collection)
    assert len(resources) == 2

    resource1 = Resource('group1', 'version1', 'plural1')
    resource2 = Resource('group2', 'version2', 'plural2')
    assert resource1 in resources
    assert resource2 in resources



import pytest

from kopf.structs.dicts import resolve


def test_existing_key():
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'hij'])
    assert r == 'val'


def test_unexisting_key_with_no_default():
    d = {'abc': {'def': {'hij': 'val'}}}
    with pytest.raises(KeyError):
        resolve(d, ['abc', 'def', 'xyz'])


def test_unexisting_key_with_default_none():
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'xyz'], None)
    assert r is None


def test_unexisting_key_with_default_value():
    default = object()
    d = {'abc': {'def': {'hij': 'val'}}}
    r = resolve(d, ['abc', 'def', 'xyz'], default)
    assert r is default


def test_nonmapping_key():
    d = {'key': 'val'}
    with pytest.raises(TypeError):
        resolve(d, ['key', 'sub'])


def test_empty_path():
    d = {'key': 'val'}
    r = resolve(d, [])
    assert r == d
    assert r is d





def test_help_in_root(invoke, mocker):
    login_pykube = mocker.patch('kopf.clients.auth.login_pykube')
    login_client = mocker.patch('kopf.clients.auth.login_client')
    verify_pykube = mocker.patch('kopf.clients.auth.verify_pykube')
    verify_client = mocker.patch('kopf.clients.auth.verify_client')

    result = invoke(['--help'])

    assert result.exit_code == 0
    assert not login_pykube.called
    assert not login_client.called
    assert not verify_pykube.called
    assert not verify_client.called

    assert 'Usage: kopf [OPTIONS]' in result.output
    assert '  run ' in result.output
    assert '  freeze ' in result.output
    assert '  resume ' in result.output


def test_help_in_subcommand(invoke, mocker):
    login_pykube = mocker.patch('kopf.clients.auth.login_pykube')
    login_client = mocker.patch('kopf.clients.auth.login_client')
    verify_pykube = mocker.patch('kopf.clients.auth.verify_pykube')
    verify_client = mocker.patch('kopf.clients.auth.verify_client')
    preload = mocker.patch('kopf.utilities.loaders.preload')
    real_run = mocker.patch('kopf.reactor.running.run')

    result = invoke(['run', '--help'])

    assert result.exit_code == 0
    assert not login_pykube.called
    assert not login_client.called
    assert not verify_pykube.called
    assert not verify_client.called
    assert not preload.called
    assert not real_run.called

    # Enough to be sure this is not a root command help.
    assert 'Usage: kopf run [OPTIONS]' in result.output
    assert '  --standalone' in result.output
    assert '  -m, --module' in result.output



"""
A registry of the handlers, attached to the resources or events.

The global registry is populated by the `kopf.on` decorators, and is used
to register the resources being watched and handled, and to attach
the handlers to the specific causes (create/update/delete/field-change).

The simple registry is part of the global registry (for each individual
resource), and also used for the sub-handlers within a top-level handler.

Both are used in the `kopf.reactor.handling` to retrieve the list
of the handlers to be executed on each reaction cycle.
"""
import abc
import functools
from types import FunctionType, MethodType
from typing import MutableMapping, NamedTuple, Text, Optional, Tuple, Callable, Mapping

from kopf.structs import filters


# An immutable reference to a custom resource definition.
class Resource(NamedTuple):
    group: Text
    version: Text
    plural: Text

    @property
    def name(self):
        return f'{self.plural}.{self.group}'

    @property
    def api_version(self):
        # Strip heading/trailing slashes if group is absent (e.g. for pods).
        return f'{self.group}/{self.version}'.strip('/')


# A registered handler (function + event meta info).
class Handler(NamedTuple):
    fn: Callable
    id: Text
    event: Text
    field: Optional[Tuple[str, ...]]
    timeout: Optional[float] = None
    initial: Optional[bool] = None
    labels: Optional[Mapping] = None
    annotations: Optional[Mapping] = None


class BaseRegistry(metaclass=abc.ABCMeta):
    """
    A registry stores the handlers and provides them to the reactor.
    """

    def get_cause_handlers(self, cause):
        return list(self._deduplicated(self.iter_cause_handlers(cause=cause)))

    @abc.abstractmethod
    def iter_cause_handlers(self, cause):
        pass

    def get_event_handlers(self, resource, event):
        return list(self._deduplicated(self.iter_event_handlers(resource=resource, event=event)))

    @abc.abstractmethod
    def iter_event_handlers(self, resource, event):
        pass

    def get_extra_fields(self, resource):
        return set(self.iter_extra_fields(resource=resource))

    @abc.abstractmethod
    def iter_extra_fields(self, resource):
        pass

    @staticmethod
    def _deduplicated(handlers):
        """
        Yield the handlers deduplicated.

        The same handler function should not be invoked more than once for one
        single event/cause, even if it is registered with multiple decorators
        (e.g. different filtering criteria or different but same-effect causes).

        One of the ways how this could happen::

            @kopf.on.create(...)
            @kopf.on.resume(...)
            def fn(**kwargs): pass

        In normal cases, the function will be called either on resource creation
        or on operator restart for the pre-existing (already handled) resources.
        When a resource is created during the operator downtime, it is
        both creation and resuming at the same time: the object is new (not yet
        handled) **AND** it is detected as per-existing before operator start.
        But `fn()` should be called only once for this cause.
        """
        seen_ids = set()
        for handler in handlers:
            if id(handler.fn) in seen_ids:
                pass
            else:
                seen_ids.add(id(handler.fn))
                yield handler


class SimpleRegistry(BaseRegistry):
    """
    A simple registry is just a list of handlers, no grouping.
    """

    def __init__(self, prefix=None):
        super().__init__()
        self.prefix = prefix
        self._handlers = []  # [Handler, ...]
        self._handlers_requiring_finalizer = []

    def __bool__(self):
        return bool(self._handlers)

    def append(self, handler):
        self._handlers.append(handler)

    def register(self, fn, id=None, event=None, field=None, timeout=None, initial=None, requires_finalizer=False,
                 labels=None, annotations=None):
        if field is None:
            field = None  # for the non-field events
        elif isinstance(field, str):
            field = tuple(field.split('.'))
        elif isinstance(field, (list, tuple)):
            field = tuple(field)
        else:
            raise ValueError(f"Field must be either a str, or a list/tuple. Got {field!r}")

        id = id if id is not None else get_callable_id(fn)
        id = id if field is None else f'{id}/{".".join(field)}'
        id = id if self.prefix is None else f'{self.prefix}/{id}'
        handler = Handler(id=id, fn=fn, event=event, field=field, timeout=timeout, initial=initial,
                          labels=labels, annotations=annotations)

        self.append(handler)

        if requires_finalizer:
            self._handlers_requiring_finalizer.append(handler)

        return fn  # to be usable as a decorator too.

    def iter_cause_handlers(self, cause):
        changed_fields = {field for _, field, _, _ in cause.diff or []}
        for handler in self._handlers:
            if handler.event is None or handler.event == cause.event:
                if handler.initial and not cause.initial:
                    pass  # ignore initial handlers in non-initial causes.
                elif filters.match(handler=handler, body=cause.body, changed_fields=changed_fields):
                    yield handler

    def iter_event_handlers(self, resource, event):
        for handler in self._handlers:
            if filters.match(handler=handler, body=event['object']):
                yield handler

    def iter_extra_fields(self, resource):
        for handler in self._handlers:
            if handler.field:
                yield handler.field

    def requires_finalizer(self, resource, body):
        # check whether the body matches a deletion handler
        for handler in self._handlers_requiring_finalizer:
            if filters.match(handler=handler, body=body):
                return True

        return False


def get_callable_id(c):
    """ Get an reasonably good id of any commonly used callable. """
    if c is None:
        return None
    elif isinstance(c, functools.partial):
        return get_callable_id(c.func)
    elif hasattr(c, '__wrapped__'):  # @functools.wraps()
        return get_callable_id(c.__wrapped__)
    elif isinstance(c, FunctionType) and c.__name__ == '<lambda>':
        # The best we can do to keep the id stable across the process restarts,
        # assuming at least no code changes. The code changes are not detectable.
        line = c.__code__.co_firstlineno
        path = c.__code__.co_filename
        return f'lambda:{path}:{line}'
    elif isinstance(c, (FunctionType, MethodType)):
        return getattr(c, '__qualname__', getattr(c, '__name__', repr(c)))
    else:
        raise ValueError(f"Cannot get id of {c!r}.")


class GlobalRegistry(BaseRegistry):
    """
    A global registry is used for handling of the multiple resources.
    It is usually populated by the `@kopf.on...` decorators.
    """

    def __init__(self):
        super().__init__()
        self._cause_handlers: MutableMapping[Resource, SimpleRegistry] = {}
        self._event_handlers: MutableMapping[Resource, SimpleRegistry] = {}

    def register_cause_handler(self, group, version, plural, fn,
                               id=None, event=None, field=None, timeout=None, initial=None, requires_finalizer=False,
                               labels=None, annotations=None):
        """
        Register an additional handler function for the specific resource and specific event.
        """
        resource = Resource(group, version, plural)
        registry = self._cause_handlers.setdefault(resource, SimpleRegistry())
        registry.register(event=event, field=field, fn=fn, id=id, timeout=timeout, initial=initial, requires_finalizer=requires_finalizer,
                          labels=labels, annotations=annotations)
        return fn  # to be usable as a decorator too.

    def register_event_handler(self, group, version, plural, fn, id=None, labels=None,
                               annotations=None):
        """
        Register an additional handler function for low-level events.
        """
        resource = Resource(group, version, plural)
        registry = self._event_handlers.setdefault(resource, SimpleRegistry())
        registry.register(fn=fn, id=id, labels=labels, annotations=annotations)
        return fn  # to be usable as a decorator too.

    @property
    def resources(self):
        """ All known resources in the registry. """
        return frozenset(self._cause_handlers) | frozenset(self._event_handlers)

    def has_cause_handlers(self, resource):
        resource_registry = self._cause_handlers.get(resource, None)
        return bool(resource_registry)

    def has_event_handlers(self, resource):
        resource_registry = self._event_handlers.get(resource, None)
        return bool(resource_registry)

    def iter_cause_handlers(self, cause):
        """
        Iterate all handlers that match this cause/event, in the order they were registered (even if mixed).
        """
        resource_registry = self._cause_handlers.get(cause.resource, None)
        if resource_registry is not None:
            yield from resource_registry.iter_cause_handlers(cause=cause)

    def iter_event_handlers(self, resource, event):
        """
        Iterate all handlers for the low-level events.
        """
        resource_registry = self._event_handlers.get(resource, None)
        if resource_registry is not None:
            yield from resource_registry.iter_event_handlers(resource=resource, event=event)

    def iter_extra_fields(self, resource):
        resource_registry = self._cause_handlers.get(resource, None)
        if resource_registry is not None:
            yield from resource_registry.iter_extra_fields(resource=resource)

    def requires_finalizer(self, resource, body):
        """
        Return whether a finalizer should be added to
        the given resource or not.
        """
        resource_registry = self._cause_handlers.get(resource, None)
        if resource_registry is None:
            return False
        return resource_registry.requires_finalizer(resource, body)


_default_registry = GlobalRegistry()


def get_default_registry() -> GlobalRegistry:
    """
    Get the default registry to be used by the decorators and the reactor
    unless the explicit registry is provided to them.
    """
    return _default_registry


def set_default_registry(registry: GlobalRegistry):
    """
    Set the default registry to be used by the decorators and the reactor
    unless the explicit registry is provided to them.
    """
    global _default_registry
    _default_registry = registry



"""
All the functions to calculate the diffs of the dicts.
"""
import collections.abc

from typing import Any, Tuple, NewType, Generator, Sequence

DiffOp = NewType('DiffOp', str)
DiffPath = Tuple[str, ...]
DiffItem = Tuple[DiffOp, DiffPath, Any, Any]
Diff = Sequence[DiffItem]


def reduce_iter(d: Diff, path: DiffPath) -> Generator[DiffItem, None, None]:
    for op, field, old, new in d:
        if not path or tuple(field[:len(path)]) == tuple(path):
            yield (op, tuple(field[len(path):]), old, new)


def reduce(d: Diff, path: DiffPath) -> Diff:
    return tuple(reduce_iter(d, path))


def diff_iter(a: Any, b: Any, path: DiffPath = ()) -> Generator[DiffItem, None, None]:
    """
    Calculate the diff between two dicts.

    Yields the tuple of form ``(op, path, old, new)``,
    where ``op`` is either ``"add"``/``"change"``/``"remove"``,
    ``path`` is a tuple with the field names (empty tuple means root),
    and the ``old`` & ``new`` values (`None` for addition/removal).

    List values are treated as a whole, and not recursed into.
    Therefore, an addition/removal of a list item is considered
    as a change of the whole value.

    If the deep diff for lists/sets is needed, see the libraries:

    * https://dictdiffer.readthedocs.io/en/latest/
    * https://github.com/seperman/deepdiff
    * https://python-json-patch.readthedocs.io/en/latest/tutorial.html
    """
    if type(a) != type(b):
        yield ('change', path, a, b)
    elif a == b:
        pass  # to exclude the case as soon as possible
    elif isinstance(a, collections.abc.Mapping):
        a_keys = frozenset(a.keys())
        b_keys = frozenset(b.keys())
        for key in b_keys - a_keys:
            yield ('add', path+(key,), None, b[key])
        for key in a_keys - b_keys:
            yield ('remove', path+(key,), a[key], None)
        for key in a_keys & b_keys:
            yield from diff_iter(a[key], b[key], path=path+(key,))
    else:
        yield ('change', path, a, b)


def diff(a: Any, b: Any, path: DiffPath = ()) -> Diff:
    """
    Same as `diff`, but returns the whole tuple instead of iterator.
    """
    return tuple(diff_iter(a, b, path=path))



"""
**THIS MODULE IS DEPRECATED AND WILL BE REMOVED.**
"""
import warnings

from kopf.engines.posting import (
    event,
    info,
    warn,
    exception,
)

__all__ = ['event', 'info', 'warn', 'exception']


# Triggered on explicit `import kopf.events` (not imported this way normally).
warnings.warn(
    "`kopf.events` is deprecated; "
    "use `kopf` directly: e.g. `kopf.event(...)`.",
    DeprecationWarning, stacklevel=0)



import kopf


def test_getting_default_registry():
    registry = kopf.get_default_registry()
    assert isinstance(registry, kopf.GlobalRegistry)


def test_setting_default_registry():
    registry_expected = kopf.GlobalRegistry()
    kopf.set_default_registry(registry_expected)
    registry_actual = kopf.get_default_registry()
    assert registry_actual is registry_expected



"""
Only the tests for async-streaming events from the sync-API calls.

Excluded: the watching routines and edge case handling, such as ERROR events
(see ``test_watching.py``).

Excluded: the queueing routines, including multiplexing and workers/handlers
(see ``test_queueing.py``).
"""

import collections.abc

import pytest

from kopf.clients.watching import StopStreaming, streaming_next, streaming_aiter


async def test_streaming_next_never_ends_with_stopiteration():
    lst = []
    src = iter(lst)

    with pytest.raises(StopStreaming) as e:
        streaming_next(src)

    assert not isinstance(e, StopIteration)
    assert not isinstance(e, StopAsyncIteration)


async def test_streaming_next_yields_and_ends():
    lst = [1, 2, 3]
    src = iter(lst)

    val1 = streaming_next(src)
    val2 = streaming_next(src)
    val3 = streaming_next(src)
    assert val1 == 1
    assert val2 == 2
    assert val3 == 3

    with pytest.raises(StopStreaming):
        streaming_next(src)


async def test_streaming_iterator_with_regular_next_yields_and_ends():
    lst = [1, 2, 3]
    src = iter(lst)

    itr = streaming_aiter(src)
    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    val1 = next(src)
    val2 = next(src)
    val3 = next(src)
    assert val1 == 1
    assert val2 == 2
    assert val3 == 3

    with pytest.raises(StopIteration):
        next(src)


async def test_streaming_iterator_with_asyncfor_works():
    lst = [1, 2, 3]
    src = iter(lst)

    itr = streaming_aiter(src)
    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    vals = []
    async for val in itr:
        vals.append(val)
    assert vals == lst


async def test_streaming_iterator_with_syncfor_fails():
    lst = [1, 2, 3]
    src = iter(lst)

    itr = streaming_aiter(src)
    assert isinstance(itr, collections.abc.AsyncIterator)
    assert isinstance(itr, collections.abc.AsyncGenerator)

    with pytest.raises(TypeError):
        for _ in itr:
            pass



"""
The main Kopf module for all the exported functions & classes.
"""

# Unlike all other places, where we import other modules and refer
# the functions via the modules, this is the framework's top-level interface,
# as it is seen by the users. So, we export the individual functions.

from kopf import (
    on,  # as a separate name on the public namespace
)
from kopf.clients.auth import (
    login,
    LoginError,
)
from kopf.config import (
    configure,
    LOGLEVEL_INFO,
    LOGLEVEL_WARNING,
    LOGLEVEL_ERROR,
    LOGLEVEL_CRITICAL,
    EventsConfig,
    WorkersConfig,
)
from kopf.engines.posting import (
    event,
    info,
    warn,
    exception,
)
from kopf.on import (
    register,
)
from kopf.reactor import (
    lifecycles,  # as a separate name on the public namespace
)
from kopf.reactor.handling import (
    TemporaryError,
    PermanentError,
    HandlerTimeoutError,
    execute,
)
from kopf.reactor.lifecycles import (
    get_default_lifecycle,
    set_default_lifecycle,
)
from kopf.reactor.registries import (
    BaseRegistry,
    SimpleRegistry,
    GlobalRegistry,
    get_default_registry,
    set_default_registry,
)
from kopf.reactor.running import (
    spawn_tasks,
    run_tasks,
    operator,
    run,
    create_tasks,  # deprecated
)
from kopf.structs.hierarchies import (
    adopt,
    label,
    harmonize_naming,
    adjust_namespace,
    build_object_reference,
    build_owner_reference,
    append_owner_reference,
    remove_owner_reference,
)

HandlerFatalError = PermanentError  # a backward-compatibility alias
HandlerRetryError = TemporaryError  # a backward-compatibility alias

__all__ = [
    'on', 'lifecycles', 'register', 'execute',
    'configure',
    'login', 'LoginError',
    'event', 'info', 'warn', 'exception',
    'spawn_tasks', 'run_tasks', 'operator', 'run', 'create_tasks',
    'adopt', 'label',
    'get_default_lifecycle', 'set_default_lifecycle',
    'build_object_reference', 'build_owner_reference',
    'append_owner_reference', 'remove_owner_reference',
    'PermanentError', 'HandlerFatalError',
    'TemporaryError', 'HandlerRetryError',
    'HandlerTimeoutError',
    'BaseRegistry',
    'SimpleRegistry',
    'GlobalRegistry',
    'get_default_registry',
    'set_default_registry',
]



"""
Kubernetes watching/streaming and the per-object queueing system.

The framework can handle multiple resources at once.
Every custom resource type is "watched" (as in ``kubectl get --watch``)
in a separate asyncio task in the never-ending loop.

The events for this resource type (of all its objects) are then pushed
to the per-object queues, which are created and destroyed dynamically.
The per-object queues are created on demand.

Every object is identified by its uid, and is handled sequentially:
i.e. the low-level events are processed in the order of their arrival.
Other objects are handled in parallel in their own sequential tasks.

To prevent the memory leaks over the long run, the queues and the workers
of each object are destroyed if no new events arrive for some time.
The destruction delay (usually few seconds, maybe minutes) is needed
to prevent the often queue/worker destruction and re-creation
in case the events are for any reason delayed by Kubernetes.

The conversion of the low-level watch-events to the high-level causes
is done in the `kopf.reactor.handling` routines.
"""

import asyncio
import logging
import time
from typing import Callable, Tuple, Union, MutableMapping, NewType, NamedTuple

import aiojobs

from kopf import config
from kopf.clients import watching
from kopf.reactor import registries

logger = logging.getLogger(__name__)


class Stream(NamedTuple):
    """ A single object's stream of watch-events, with some extra helpers. """
    watchevents: asyncio.Queue
    replenished: asyncio.Event  # means: "hurry up, there are new events queued again"


ObjectUid = NewType('ObjectUid', str)
ObjectRef = Tuple[registries.Resource, ObjectUid]
Streams = MutableMapping[ObjectRef, Stream]

EOS = object()
""" An end-of-stream marker sent from the watcher to the workers. """


# TODO: add the label_selector support for the dev-mode?
async def watcher(
        namespace: Union[None, str],
        resource: registries.Resource,
        handler: Callable,
):
    """
    The watchers watches for the resource events via the API, and spawns the handlers for every object.

    All resources and objects are done in parallel, but one single object is handled sequentially
    (otherwise, concurrent handling of multiple events of the same object could cause data damage).

    The watcher is as non-blocking and async, as possible. It does neither call any external routines,
    nor it makes the API calls via the sync libraries.

    The watcher is generally a never-ending task (unless an error happens or it is cancelled).
    The workers, on the other hand, are limited approximately to the life-time of an object's event.
    """

    # All per-object workers are handled as fire-and-forget jobs via the scheduler,
    # and communicated via the per-object event queues.
    scheduler = await aiojobs.create_scheduler(limit=config.WorkersConfig.queue_workers_limit)
    streams: Streams = {}
    try:
        # Either use the existing object's queue, or create a new one together with the per-object job.
        # "Fire-and-forget": we do not wait for the result; the job destroys itself when it is fully done.
        async for event in watching.infinite_watch(resource=resource, namespace=namespace):
            key = (resource, event['object']['metadata']['uid'])
            try:
                streams[key].replenished.set()  # interrupt current sleeps, if any.
                await streams[key].watchevents.put(event)
            except KeyError:
                streams[key] = Stream(watchevents=asyncio.Queue(), replenished=asyncio.Event())
                streams[key].replenished.set()  # interrupt current sleeps, if any.
                await streams[key].watchevents.put(event)
                await scheduler.spawn(worker(handler=handler, streams=streams, key=key))

        # Allow the existing workers to finish gracefully before killing them.
        await _wait_for_depletion(scheduler=scheduler, streams=streams)

    finally:
        # Forcedly terminate all the fire-and-forget per-object jobs, of they are still running.
        await asyncio.shield(scheduler.close())


async def worker(
        handler: Callable,
        streams: Streams,
        key: ObjectRef,
):
    """
    The per-object workers consume the object's events and invoke the handler.

    The handler is expected to be an async coroutine, always the one from the framework.
    In fact, it is either a peering handler, which monitors the peer operators,
    or a generic resource handler, which internally calls the registered synchronous handlers.

    The per-object worker is a time-limited task, which ends as soon as all the object's events
    have been handled. The watcher will spawn a new job when and if the new events arrive.

    To prevent the queue/job deletion and re-creation to happen too often, the jobs wait some
    reasonable, but small enough time (few seconds) before actually finishing --
    in case the new events are there, but the API or the watcher task lags a bit.
    """
    watchevents = streams[key].watchevents
    replenished = streams[key].replenished
    shouldstop = False
    try:
        while not shouldstop:

            # Try ASAP, but give it few seconds for the new events to arrive, maybe.
            # If the queue is empty for some time, then indeed finish the object's worker.
            # If the queue is filled, use the latest event only (within the short timeframe).
            # If an EOS marker is received, handle the last real event, then finish the worker ASAP.
            try:
                event = await asyncio.wait_for(
                    watchevents.get(),
                    timeout=config.WorkersConfig.worker_idle_timeout)
            except asyncio.TimeoutError:
                break
            else:
                try:
                    while True:
                        prev_event = event
                        next_event = await asyncio.wait_for(
                            watchevents.get(),
                            timeout=config.WorkersConfig.worker_batch_window)
                        shouldstop = shouldstop or next_event is EOS
                        event = prev_event if next_event is EOS else next_event
                except asyncio.TimeoutError:
                    pass

            # Exit gracefully and immediately on the end-of-stream marker sent by the watcher.
            if event is EOS:
                break

            # Try the handler. In case of errors, show the error, but continue the queue processing.
            replenished.clear()
            try:
                await handler(event=event, replenished=replenished)
            except Exception as e:
                # TODO: handler is a functools.partial. make the prints a bit nicer by removing it.
                logger.exception(f"{handler} failed with an exception. Ignoring the event.")
                # raise

    finally:
        # Whether an exception or a break or a success, notify the caller, and garbage-collect our queue.
        # The queue must not be left in the queue-cache without a corresponding job handling this queue.
        try:
            del streams[key]
        except KeyError:
            pass


async def _wait_for_depletion(*, scheduler: aiojobs.Scheduler, streams: Streams):

    # Notify all the workers to finish now. Wake them up if they are waiting in the queue-getting.
    for stream in streams.values():
        await stream.watchevents.put(EOS)

    # Wait for the queues to be depleted, but only if there are some workers running.
    # Continue with the tasks termination if the timeout is reached, no matter the queues.
    started = time.perf_counter()
    while streams and \
            scheduler.active_count and \
            time.perf_counter() - started < config.WorkersConfig.worker_exit_timeout:
        await asyncio.sleep(config.WorkersConfig.worker_exit_timeout / 100.)

    # The last check if the termination is going to be graceful or not.
    if streams:
        logger.warning("Unprocessed streams left for %r.", list(streams.keys()))



"""
All the functions to properly build the object hierarchies.
"""
from kopf.structs import dicts


def build_object_reference(body):
    """
    Construct an object reference for the events.

    Keep in mind that some fields can be absent: e.g. ``namespace``
    for cluster resources, or e.g. ``apiVersion`` for ``kind: Node``, etc.
    """
    ref = dict(
        apiVersion=body.get('apiVersion'),
        kind=body.get('kind'),
        name=body.get('metadata', {}).get('name'),
        uid=body.get('metadata', {}).get('uid'),
        namespace=body.get('metadata', {}).get('namespace'),
    )
    return {key: val for key, val in ref.items() if val}


def build_owner_reference(body):
    """
    Construct an owner reference object for the parent-children relationships.

    The structure needed to link the children objects to the current object as a parent.
    See https://kubernetes.io/docs/concepts/workloads/controllers/garbage-collection/

    Keep in mind that some fields can be absent: e.g. ``namespace``
    for cluster resources, or e.g. ``apiVersion`` for ``kind: Node``, etc.
    """
    ref = dict(
        controller=True,
        blockOwnerDeletion=True,
        apiVersion=body.get('apiVersion'),
        kind=body.get('kind'),
        name=body.get('metadata', {}).get('name'),
        uid=body.get('metadata', {}).get('uid'),
    )
    return {key: val for key, val in ref.items() if val}


def append_owner_reference(objs, owner):
    """
    Append an owner reference to the resource(s), if it is not yet there.

    Note: the owned objects are usually not the one being processed,
    so the whole body can be modified, no patches are needed.
    """
    owner = build_owner_reference(owner)
    for obj in dicts.walk(objs):
        refs = obj.setdefault('metadata', {}).setdefault('ownerReferences', [])
        matching = [ref for ref in refs if ref.get('uid') == owner.get('uid')]
        if not matching:
            refs.append(owner)


def remove_owner_reference(objs, owner):
    """
    Remove an owner reference to the resource(s), if it is there.

    Note: the owned objects are usually not the one being processed,
    so the whole body can be modified, no patches are needed.
    """
    owner = build_owner_reference(owner)
    for obj in dicts.walk(objs):
        refs = obj.setdefault('metadata', {}).setdefault('ownerReferences', [])
        matching = [ref for ref in refs if ref.get('uid') == owner.get('uid')]
        for ref in matching:
            refs.remove(ref)


def label(objs, labels, *, force=False, nested=None):
    """
    Apply the labels to the object(s).
    """
    for obj in dicts.walk(objs, nested=nested):
        obj_labels = obj.setdefault('metadata', {}).setdefault('labels', {})
        for key, val in labels.items():
            if force:
                obj_labels[key] = val
            else:
                obj_labels.setdefault(key, val)


def harmonize_naming(objs, name=None, strict=False):
    """
    Adjust the names or prefixes of the objects.

    In strict mode, the provided name is used as is. It can be helpful
    if the object is referred by that name in other objects.

    In non-strict mode (the default), the object uses the provided name
    as a prefix, while the suffix is added by Kubernetes remotely.
    The actual name should be taken from Kubernetes response
    (this is the recommended scenario).

    If the objects already have their own names, auto-naming is not applied,
    and the existing names are used as is.
    """
    for obj in dicts.walk(objs):
        if obj.get('metadata', {}).get('name', None) is None:
            if strict:
                obj.setdefault('metadata', {}).setdefault('name', name)
            else:
                obj.setdefault('metadata', {}).setdefault('generateName', f'{name}-')


def adjust_namespace(objs, namespace=None):
    """
    Adjust the namespace of the objects.

    If the objects already have the namespace set, it will be preserved.

    It is a common practice to keep the children objects in the same
    namespace as their owner, unless explicitly overridden at time of creation.
    """
    for obj in dicts.walk(objs):
        obj.setdefault('metadata', {}).setdefault('namespace', namespace)


def adopt(objs, owner, *, nested=None):
    """
    The children should be in the same namespace, named after their parent, and owned by it.
    """
    append_owner_reference(objs, owner=owner)
    harmonize_naming(objs, name=owner.get('metadata', {}).get('name', None))
    adjust_namespace(objs, namespace=owner.get('metadata', {}).get('namespace', None))
    label(objs, labels=owner.get('metadata', {}).get('labels', {}), nested=nested)



"""
Detection of the event causes, based on the resource state.

The low-level watch-events are highly limited in information on what
caused them, and they only notify that the object was changed somehow:

* ``ADDED`` for the newly created objects (or for the first-time listing).
* ``MODIFIED`` for the changes of any field, be that metainfo, spec, or status.
* ``DELETED`` for the actual deletion of the object post-factum.

The conversion of low-level *events* to high level *causes* is done by
checking the object's state and comparing it to the saved last-seen state.

This allows to track which specific fields were changed, and if are those
changes are important enough to call the handlers: e.g. the ``status`` changes
are ignored, so as some selected system fields of the ``metainfo``.

For deletion, the cause is detected when the object is just marked for deletion,
not when it is actually deleted (as the events notify): so that the handlers
could execute on the yet-existing object (and its children, if created).
"""
import logging
from typing import NamedTuple, Text, Mapping, MutableMapping, Optional, Any, Union

from kopf.reactor import registries
from kopf.structs import diffs
from kopf.structs import finalizers
from kopf.structs import lastseen

# Constants for event types, to prevent a direct usage of strings, and typos.
# They are not exposed by the framework, but are used internally. See also: `kopf.on`.
CREATE = 'create'
UPDATE = 'update'
DELETE = 'delete'
RESUME = 'resume'
NOOP = 'noop'
FREE = 'free'
GONE = 'gone'
ACQUIRE = 'acquire'
RELEASE = 'release'

# These sets are checked in few places, so we keep them centralised:
# the user-facing causes (for handlers), and internally facing (so as handled).
HANDLER_CAUSES = (CREATE, UPDATE, DELETE, RESUME)
REACTOR_CAUSES = (NOOP, FREE, GONE, ACQUIRE, RELEASE)
ALL_CAUSES = HANDLER_CAUSES + REACTOR_CAUSES

# The human-readable names of these causes. Will be capitalised when needed.
TITLES = {
    CREATE: 'creation',
    UPDATE: 'update',
    DELETE: 'deletion',
    RESUME: 'resuming',
}


class Cause(NamedTuple):
    """
    The cause is what has caused the whole reaction as a chain of handlers.

    Unlike the low-level Kubernetes watch-events, the cause is aware
    of actual field changes, including multi-handler changes.
    """
    logger: Union[logging.Logger, logging.LoggerAdapter]
    resource: registries.Resource
    event: Text
    initial: bool
    body: MutableMapping
    patch: MutableMapping
    digest: Union[str, int]
    diff: Optional[diffs.Diff] = None
    old: Optional[Any] = None
    new: Optional[Any] = None


def detect_cause(
        event: Mapping,
        requires_finalizer: bool = True,
        **kwargs
) -> Cause:
    """
    Detect the cause of the event to be handled.

    This is a purely computational function with no side-effects.
    The causes are then consumed by `custom_object_handler`,
    which performs the actual handler invocation, logging, patching,
    and other side-effects.
    """
    diff = kwargs.get('diff')
    body = event['object']
    initial = event['type'] is None  # special value simulated by us in kopf.reactor.watching.

    # The object was really deleted from the cluster. But we do not care anymore.
    if event['type'] == 'DELETED':
        return Cause(
            event=GONE,
            body=body,
            initial=initial,
            **kwargs)

    # The finalizer has been just removed. We are fully done.
    if finalizers.is_deleted(body) and not finalizers.has_finalizers(body):
        return Cause(
            event=FREE,
            body=body,
            initial=initial,
            **kwargs)

    # Marked for deletion, but we still hold it with our finalizer.
    if finalizers.is_deleted(body):
        return Cause(
            event=DELETE,
            body=body,
            initial=initial,
            **kwargs)

    # For a fresh new object, first block it from accidental deletions without our permission.
    # The actual handler will be called on the next call.
    # Only return this cause if the resource requires finalizers to be added.
    if requires_finalizer and not finalizers.has_finalizers(body):
        return Cause(
            event=ACQUIRE,
            body=body,
            initial=initial,
            **kwargs)

    # Check whether or not the resource has finalizers, but doesn't require them. If this is
    # the case, then a resource may not be able to be deleted completely as finalizers may
    # not be removed by the operator under normal operation. We remove the finalizers first,
    # and any handler that should be called will be done on the next call.
    if not requires_finalizer and finalizers.has_finalizers(body):
        return Cause(
            event=RELEASE,
            body=body,
            initial=initial,
            **kwargs)

    # For an object seen for the first time (i.e. just-created), call the creation handlers,
    # then mark the state as if it was seen when the creation has finished.
    if not lastseen.has_state(body):
        kwargs['digest'] = True  # or any other true'ish constant (str/int)
        return Cause(
            event=CREATE,
            body=body,
            initial=initial,
            **kwargs)

    # Cases with no state changes are usually ignored (NOOP). But for the "None" events,
    # as simulated for the initial listing, we call the resuming handlers (e.g. threads/tasks).
    if not diff and initial:
        return Cause(
            event=RESUME,
            body=body,
            initial=initial,
            **kwargs)

    # The previous step triggers one more patch operation without actual changes. Ignore it.
    # Either the last-seen state or the status field has changed.
    if not diff:
        return Cause(
            event=NOOP,
            body=body,
            initial=initial,
            **kwargs)

    # And what is left, is the update operation on one of the useful fields of the existing object.
    return Cause(
        event=UPDATE,
        body=body,
        initial=initial,
        **kwargs)


def enrich_cause(
        cause: Cause,
        **kwargs
) -> Cause:
    """
    Produce a new derived cause with some fields modified ().

    Usually, those are the old/new/diff fields, and used when a field-handler
    is invoked (the old/new/diff refer to the field's values only).
    """
    return cause._replace(**kwargs)



"""
Advanced modes of sleeping.
"""
import asyncio
from typing import Optional


async def sleep_or_wait(
        delay: float,
        event: asyncio.Event,
) -> Optional[float]:
    """
    Measure the sleep time: either until the timeout, or until the event is set.

    Returns the number of seconds left to sleep, or ``None`` if the sleep was
    not interrupted and reached its specified delay (an equivalent of ``0``).
    In theory, the result can be ``0`` if the sleep was interrupted precisely
    the last moment before timing out; this is unlikely to happen though.
    """
    loop = asyncio.get_running_loop()
    try:
        start_time = loop.time()
        await asyncio.wait_for(event.wait(), timeout=delay)
    except asyncio.TimeoutError:
        return None  # interruptable sleep is over: uninterrupted.
    else:
        end_time = loop.time()
        duration = end_time - start_time
        return max(0, delay - duration)



import pytest

import kopf
from kopf.reactor.causation import Cause
from kopf.reactor.invocation import invoke


@pytest.mark.parametrize('lifecycle', [
    kopf.lifecycles.all_at_once,
    kopf.lifecycles.one_by_one,
    kopf.lifecycles.randomized,
    kopf.lifecycles.shuffled,
    kopf.lifecycles.asap,
])
async def test_protocol_invocation(mocker, lifecycle):
    """
    To be sure that all kwargs are accepted properly.
    Especially when the new kwargs are added or an invocation protocol changed.
    """
    cause = mocker.Mock(spec=Cause)
    handlers = []
    selected = await invoke(lifecycle, handlers, cause=cause)
    assert isinstance(selected, (tuple, list))
    assert len(selected) == 0



import pytest
import requests

from kopf.clients.fetching import list_objs


def test_when_successful_clustered(req_mock, resource):
    result = {'items': []}
    req_mock.get.return_value.json.return_value = result

    lst = list_objs(resource=resource, namespace=None)
    assert lst is result

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/kopfexamples' in url
    assert 'namespaces/' not in url


def test_when_successful_namespaced(req_mock, resource):
    result = {'items': []}
    req_mock.get.return_value.json.return_value = result

    lst = list_objs(resource=resource, namespace='ns1')
    assert lst is result

    assert req_mock.get.called
    assert req_mock.get.call_count == 1

    url = req_mock.get.call_args_list[0][1]['url']
    assert 'apis/zalando.org/v1/namespaces/ns1/kopfexamples' in url


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 403, 500, 666])
def test_raises_api_error(req_mock, resource, namespace, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.get.side_effect = error

    with pytest.raises(requests.exceptions.HTTPError) as e:
        list_objs(resource=resource, namespace=namespace)
    assert e.value.response.status_code == status



import asyncio

import pytest
from asynctest import call

from kopf import event, info, warn, exception
from kopf.engines.posting import poster, K8sEvent, event_queue_var, event_queue_loop_var

OBJ1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}}
REF1 = {'apiVersion': 'group1/version1', 'kind': 'Kind1',
        'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1'}
OBJ2 = {'apiVersion': 'group2/version2', 'kind': 'Kind2',
        'metadata': {'uid': 'uid2', 'name': 'name2', 'namespace': 'ns2'}}
REF2 = {'apiVersion': 'group2/version2', 'kind': 'Kind2',
        'uid': 'uid2', 'name': 'name2', 'namespace': 'ns2'}


async def test_poster_polls_and_posts(mocker):
    event1 = K8sEvent(type='type1', reason='reason1', message='message1', ref=REF1)
    event2 = K8sEvent(type='type2', reason='reason2', message='message2', ref=REF2)
    event_queue = asyncio.Queue()
    event_queue.put_nowait(event1)
    event_queue.put_nowait(event2)

    # A way to cancel `while True` cycle when we need it (ASAP).
    def _cancel(*args, **kwargs):
        if post_event.call_count >= 2:
            raise asyncio.CancelledError()
    post_event = mocker.patch('kopf.clients.events.post_event', side_effect=_cancel)

    # A way to cancel `whole True` cycle by timing, event if routines are not called.
    with pytest.raises(asyncio.CancelledError):
        await asyncio.wait_for(
            poster(event_queue=event_queue), timeout=0.5)

    assert post_event.call_count == 2
    assert post_event.await_count == 2
    assert post_event.called_with(
        call(ref=REF1, type='type1', reason='reason1', message='message1'),
        call(ref=REF2, type='type2', reason='reason2', message='message2'),
    )


def test_queueing_fails_with_no_queue(event_queue_loop):
    # Prerequisite: the context-var should not be set by anything in advance.
    sentinel = object()
    assert event_queue_var.get(sentinel) is sentinel

    with pytest.raises(LookupError):
        event(OBJ1, type='type1', reason='reason1', message='message1')


def test_queueing_fails_with_no_loop(event_queue):
    # Prerequisite: the context-var should not be set by anything in advance.
    sentinel = object()
    assert event_queue_loop_var.get(sentinel) is sentinel

    with pytest.raises(LookupError):
        event(OBJ1, type='type1', reason='reason1', message='message1')


async def test_via_event_function(mocker, event_queue, event_queue_loop):
    post_event = mocker.patch('kopf.clients.events.post_event')

    event(OBJ1, type='type1', reason='reason1', message='message1')

    assert not post_event.called
    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()

    assert isinstance(event1, K8sEvent)
    assert event1.ref == REF1
    assert event1.type == 'type1'
    assert event1.reason == 'reason1'
    assert event1.message == 'message1'


@pytest.mark.parametrize('event_fn, event_type', [
    pytest.param(info, "Normal", id='info'),
    pytest.param(warn, "Warning", id='warn'),
    pytest.param(exception, "Error", id='exception'),
])
async def test_via_shortcut(mocker, event_fn, event_type, event_queue, event_queue_loop):
    post_event = mocker.patch('kopf.clients.events.post_event')

    event_fn(OBJ1, reason='reason1', message='message1')

    assert not post_event.called
    assert event_queue.qsize() == 1
    event1 = event_queue.get_nowait()

    assert isinstance(event1, K8sEvent)
    assert event1.ref == REF1
    assert event1.type == event_type
    assert event1.reason == 'reason1'
    assert event1.message == 'message1'



import kopf


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': 'somevalue'})
def create_with_labels_satisfied(logger, **kwargs):
    logger.info("Label satisfied.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': None})
def create_with_labels_exist(logger, **kwargs):
    logger.info("Label exists.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', labels={'somelabel': 'othervalue'})
def create_with_labels_not_satisfied(logger, **kwargs):
    logger.info("Label not satisfied.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': 'somevalue'})
def create_with_annotations_satisfied(logger, **kwargs):
    logger.info("Annotation satisfied.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': None})
def create_with_annotations_exist(logger, **kwargs):
    logger.info("Annotation exists.")


@kopf.on.create('zalando.org', 'v1', 'kopfexamples', annotations={'someannotation': 'othervalue'})
def create_with_annotations_not_satisfied(logger, **kwargs):
    logger.info("Annotation not satisfied.")



"""
A connection between object logger and background k8s-event poster.

Everything logged to the object logger (except for debug information)
is also posted as the k8s-event -- in the background: `kopf.engines.posting`.

This eliminates the need to log & post the same messages, which complicates
the operators' code, and can lead to information loss or mismatch
(e.g. when logging call is added, but posting is forgotten).
"""
import copy
import logging

from kopf.engines import posting


class ObjectPrefixingFormatter(logging.Formatter):
    """ An utility to prefix the per-object log messages. """

    def format(self, record):
        if hasattr(record, 'k8s_ref'):
            ref = record.k8s_ref
            prefix = f"[{ref.get('namespace', '')}/{ref.get('name', '')}]"
            record = copy.copy(record)  # shallow
            record.msg = f"{prefix} {record.msg}"
        return super().format(record)


class K8sPoster(logging.Handler):
    """
    A handler to post all log messages as K8s events.
    """

    def createLock(self):
        # Save some time on unneeded locks. Events are posted in the background.
        # We only put events to the queue, which is already lock-protected.
        self.lock = None

    def filter(self, record):
        # Only those which have a k8s object referred (see: `ObjectLogger`).
        # Otherwise, we have nothing to post, and nothing to do.
        has_ref = hasattr(record, 'k8s_ref')
        skipped = hasattr(record, 'k8s_skip') and record.k8s_skip
        return has_ref and not skipped and super().filter(record)

    def emit(self, record):
        # Same try-except as in e.g. `logging.StreamHandler`.
        try:
            type = (
                "Debug" if record.levelno <= logging.DEBUG else
                "Normal" if record.levelno <= logging.INFO else
                "Warning" if record.levelno <= logging.WARNING else
                "Error" if record.levelno <= logging.ERROR else
                "Fatal" if record.levelno <= logging.FATAL else
                logging.getLevelName(record.levelno).capitalize())
            reason = 'Logging'
            message = self.format(record)
            posting.enqueue(
                ref=record.k8s_ref,
                type=type,
                reason=reason,
                message=message)
        except Exception:
            self.handleError(record)


class ObjectLogger(logging.LoggerAdapter):
    """
    A logger/adapter to carry the object identifiers for formatting.

    The identifiers are then used both for formatting the per-object messages
    in `ObjectPrefixingFormatter`, and when posting the per-object k8s-events.

    Constructed in event handling of each individual object.

    The internal structure is made the same as an object reference in K8s API,
    but can change over time to anything needed for our internal purposes.
    However, as little information should be carried as possible,
    and the information should be protected against the object modification
    (e.g. in case of background posting via the queue; see `K8sPoster`).
    """

    def __init__(self, *, body):
        super().__init__(logger, dict(
            k8s_skip=False,
            k8s_ref=dict(
                apiVersion=body.get('apiVersion'),
                kind=body.get('kind'),
                name=body.get('metadata', {}).get('name'),
                uid=body.get('metadata', {}).get('uid'),
                namespace=body.get('metadata', {}).get('namespace'),
            ),
        ))

    def process(self, msg, kwargs):
        # Native logging overwrites the message's extra with the adapter's extra.
        # We merge them, so that both message's & adapter's extras are available.
        kwargs["extra"] = dict(self.extra, **kwargs.get('extra', {}))
        return msg, kwargs

    def log(self, level, msg, *args, local=False, **kwargs):
        if local:
            kwargs['extra'] = dict(kwargs.pop('extra', {}), k8s_skip=True)
        super().log(level, msg, *args, **kwargs)


logger = logging.getLogger('kopf.objects')
logger.addHandler(K8sPoster(level=logging.INFO))



import time

import kopf


class MyException(Exception):
    pass


@kopf.on.create('zalando.org', 'v1', 'kopfexamples')
def create_fn(retry, **kwargs):
    time.sleep(2)  # for different timestamps of the events
    if not retry:
        raise kopf.TemporaryError("First failure.", delay=10)
    elif retry == 1:
        raise MyException("Second failure.")
    else:
        raise kopf.PermanentError("Third failure, the final one.")



import re
import shlex
import subprocess
import time

import pytest

from kopf.testing import KopfRunner


def test_all_examples_are_runnable(mocker, with_crd, exampledir):

    # If the example has its own opinion on the timing, try to respect it.
    # See e.g. /examples/99-all-at-once/example.py.
    example_py = exampledir / 'example.py'
    m = re.search(r'^E2E_CREATE_TIME\s*=\s*(\d+)$', example_py.read_text(), re.M)
    e2e_create_time = eval(m.group(1)) if m else None
    m = re.search(r'^E2E_DELETE_TIME\s*=\s*(\d+)$', example_py.read_text(), re.M)
    e2e_delete_time = eval(m.group(1)) if m else None
    m = re.search(r'^E2E_TRACEBACKS\s*=\s*(\w+)$', example_py.read_text(), re.M)
    e2e_tracebacks = eval(m.group(1)) if m else None
    # check whether there are mandatory deletion handlers or not
    m = re.search(r'@kopf\.on\.delete\((\s|.*)?(optional=(\w+))?\)', example_py.read_text(), re.M)
    requires_finalizer = False
    if m:
        requires_finalizer = True
        if m.group(2):
            requires_finalizer = not eval(m.group(3))

    # Skip the e2e test if the framework-optional but test-required library is missing.
    m = re.search(r'import kubernetes', example_py.read_text(), re.M)
    if m:
        pytest.importorskip('kubernetes')

    # To prevent lengthy sleeps on the simulated retries.
    mocker.patch('kopf.reactor.handling.DEFAULT_RETRY_DELAY', 1)

    # To prevent lengthy threads in the loop executor when the process exits.
    mocker.patch('kopf.config.WatchersConfig.default_stream_timeout', 10)

    # Run an operator and simulate some activity with the operated resource.
    with KopfRunner(['run', '--standalone', '--verbose', str(example_py)], timeout=60) as runner:
        subprocess.run("kubectl apply -f examples/obj.yaml", shell=True, check=True)
        time.sleep(e2e_create_time or 2)  # give it some time to react and to sleep and to retry
        subprocess.run("kubectl delete -f examples/obj.yaml", shell=True, check=True)
        time.sleep(e2e_delete_time or 1)  # give it some time to react

    # Ensure that the operator did not die on start, or during the operation.
    assert runner.exception is None
    assert runner.exit_code == 0

    # There are usually more than these messages, but we only check for the certain ones.
    # This just shows us that the operator is doing something, it is alive.
    if requires_finalizer:
        assert '[default/kopf-example-1] Adding the finalizer' in runner.stdout
    assert '[default/kopf-example-1] Creation event:' in runner.stdout
    if requires_finalizer:
        assert '[default/kopf-example-1] Deletion event:' in runner.stdout
    assert '[default/kopf-example-1] Deleted, really deleted' in runner.stdout
    if not e2e_tracebacks:
        assert 'Traceback (most recent call last):' not in runner.stdout



import asyncio
import logging

import pytest

import kopf
from kopf.reactor.causation import ALL_CAUSES, HANDLER_CAUSES
from kopf.reactor.handling import custom_object_handler


@pytest.mark.parametrize('cause_type', ALL_CAUSES)
async def test_all_logs_are_prefixed(registry, resource, handlers,
                                     logstream, cause_type, cause_mock):
    cause_mock.event = cause_type

    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )

    lines = logstream.getvalue().splitlines()
    assert lines  # no messages means that we cannot test it
    assert all(line.startswith('prefix [ns1/name1] ') for line in lines)


@pytest.mark.parametrize('diff', [
    pytest.param((('op', ('field',), 'old', 'new'),), id='realistic-diff'),
    pytest.param((), id='empty-tuple-diff'),
    pytest.param([], id='empty-list-diff'),
])
@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_diffs_logged_if_present(registry, resource, handlers, cause_type, cause_mock,
                                       caplog, assert_logs, diff):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = cause_type
    cause_mock.diff = diff
    cause_mock.new = object()  # checked for `not None`
    cause_mock.old = object()  # checked for `not None`

    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )
    assert_logs([
        " event: ",
        " diff: "
    ])


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_diffs_not_logged_if_absent(registry, resource, handlers, cause_type, cause_mock,
                                          caplog, assert_logs):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = cause_type
    cause_mock.diff = None  # same as the default, but for clarity

    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )
    assert_logs([
        " event: ",
    ], prohibited=[
        " diff: "
    ])



import json

import pykube.exceptions
import pytest
import requests

from kopf.clients.patching import patch_obj


@pytest.mark.resource_clustered  # see `req_mock`
async def test_by_name_clustered(req_mock, resource):
    patch = {'x': 'y'}
    res = await patch_obj(resource=resource, namespace=None, name='name1', patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert req_mock.patch.called
    assert req_mock.patch.call_count == 1

    url = req_mock.patch.call_args_list[0][1]['url']
    assert 'namespaces/' not in url
    assert f'apis/{resource.api_version}/{resource.plural}/name1' in url

    data = json.loads(req_mock.patch.call_args_list[0][1]['data'])
    assert data == {'x': 'y'}


async def test_by_name_namespaced(req_mock, resource):
    patch = {'x': 'y'}
    res = await patch_obj(resource=resource, namespace='ns1', name='name1', patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert req_mock.patch.called
    assert req_mock.patch.call_count == 1

    url = req_mock.patch.call_args_list[0][1]['url']
    assert 'namespaces/' in url
    assert f'apis/{resource.api_version}/namespaces/ns1/{resource.plural}/name1' in url

    data = json.loads(req_mock.patch.call_args_list[0][1]['data'])
    assert data == {'x': 'y'}


@pytest.mark.resource_clustered  # see `req_mock`
async def test_by_body_clustered(req_mock, resource):
    patch = {'x': 'y'}
    body = {'metadata': {'name': 'name1'}}
    res = await patch_obj(resource=resource, body=body, patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert req_mock.patch.called
    assert req_mock.patch.call_count == 1

    url = req_mock.patch.call_args_list[0][1]['url']
    assert 'namespaces/' not in url
    assert f'apis/{resource.api_version}/{resource.plural}/name1' in url

    data = json.loads(req_mock.patch.call_args_list[0][1]['data'])
    assert data == {'x': 'y'}


async def test_by_body_namespaced(req_mock, resource):
    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    res = await patch_obj(resource=resource, body=body, patch=patch)
    assert res is None  # never return any k8s-client specific things

    assert req_mock.patch.called
    assert req_mock.patch.call_count == 1

    url = req_mock.patch.call_args_list[0][1]['url']
    assert 'namespaces/' in url
    assert f'apis/{resource.api_version}/namespaces/ns1/{resource.plural}/name1' in url

    data = json.loads(req_mock.patch.call_args_list[0][1]['data'])
    assert data == {'x': 'y'}


async def test_raises_when_body_conflicts_with_namespace(req_mock, resource):
    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        await patch_obj(resource=resource, body=body, namespace='ns1', patch=patch)

    assert not req_mock.patch.called


async def test_raises_when_body_conflicts_with_name(req_mock, resource):
    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        await patch_obj(resource=resource, body=body, name='name1', patch=patch)

    assert not req_mock.patch.called


async def test_raises_when_body_conflicts_with_ids(req_mock, resource):
    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(TypeError):
        await patch_obj(resource=resource, body=body, namespace='ns1', name='name1', patch=patch)

    assert not req_mock.patch.called


@pytest.mark.parametrize('status', [404])
async def test_ignores_absent_objects_with_requests_httperror(req_mock, resource, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.patch.side_effect = error

    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    await patch_obj(resource=resource, body=body, patch=patch)


@pytest.mark.parametrize('status', [404])
async def test_ignores_absent_objects_with_pykube_httperror(req_mock, resource, status):
    error = pykube.exceptions.HTTPError(status, "boo!")
    req_mock.patch.side_effect = error

    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    await patch_obj(resource=resource, body=body, patch=patch)


@pytest.mark.parametrize('namespace', [None, 'ns1'], ids=['without-namespace', 'with-namespace'])
@pytest.mark.parametrize('status', [400, 401, 403, 500, 666])
async def test_raises_api_errors(req_mock, resource, namespace, status):
    response = requests.Response()
    response.status_code = status
    error = requests.exceptions.HTTPError("boo!", response=response)
    req_mock.patch.side_effect = error

    patch = {'x': 'y'}
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    with pytest.raises(requests.exceptions.HTTPError) as e:
        await patch_obj(resource=resource, body=body, patch=patch)
    assert e.value.response.status_code == status



import asyncio
import logging

import pytest

import kopf
from kopf.reactor.causation import HANDLER_CAUSES
from kopf.reactor.handling import custom_object_handler
from kopf.structs.lastseen import LAST_SEEN_ANNOTATION


@pytest.mark.parametrize('cause_type', HANDLER_CAUSES)
async def test_skipped_with_no_handlers(
        registry, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)
    cause_mock.event = cause_type

    assert not registry.has_cause_handlers(resource=resource)  # prerequisite
    registry.register_cause_handler(
        group=resource.group,
        version=resource.version,
        plural=resource.plural,
        event='a-non-existent-cause-type',
        fn=lambda **_: None,
    )
    assert registry.has_cause_handlers(resource=resource)  # prerequisite

    await custom_object_handler(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        resource=resource,
        event={'type': 'irrelevant', 'object': cause_mock.body},
        freeze=asyncio.Event(),
        replenished=asyncio.Event(),
        event_queue=asyncio.Queue(),
    )

    assert not k8s_mocked.asyncio_sleep.called
    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    # The patch must contain ONLY the last-seen update, and nothing else.
    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert set(patch.keys()) == {'metadata'}
    assert set(patch['metadata'].keys()) == {'annotations'}
    assert set(patch['metadata']['annotations'].keys()) == {LAST_SEEN_ANNOTATION}

    assert_logs([
        ".* event:",
        "Patching with:",
    ], prohibited=[
        "All handlers succeeded",
    ])



import asyncio
import functools
import logging
import signal
import threading
import warnings
from typing import Optional, Callable, Collection

from kopf.engines import peering
from kopf.engines import posting
from kopf.reactor import handling
from kopf.reactor import lifecycles
from kopf.reactor import queueing
from kopf.reactor import registries

logger = logging.getLogger(__name__)


def run(
        loop: Optional[asyncio.AbstractEventLoop] = None,
        lifecycle: Optional[Callable] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        standalone: bool = False,
        priority: int = 0,
        peering_name: Optional[str] = None,
        namespace: Optional[str] = None,
):
    """
    Run the whole operator synchronously.

    This function should be used to run an operator in normal sync mode.
    """
    loop = loop if loop is not None else asyncio.get_event_loop()
    try:
        loop.run_until_complete(operator(
            lifecycle=lifecycle,
            registry=registry,
            standalone=standalone,
            namespace=namespace,
            priority=priority,
            peering_name=peering_name,
        ))
    except asyncio.CancelledError:
        pass


async def operator(
        lifecycle: Optional[Callable] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        standalone: bool = False,
        priority: int = 0,
        peering_name: Optional[str] = None,
        namespace: Optional[str] = None,
):
    """
    Run the whole operator asynchronously.

    This function should be used to run an operator in an asyncio event-loop
    if the operator is orchestrated explicitly and manually.

    It is efficiently `spawn_tasks` + `run_tasks` with some safety.
    """
    existing_tasks = await _all_tasks()
    operator_tasks = await spawn_tasks(
        lifecycle=lifecycle,
        registry=registry,
        standalone=standalone,
        namespace=namespace,
        priority=priority,
        peering_name=peering_name,
    )
    await run_tasks(operator_tasks, ignored=existing_tasks)


async def spawn_tasks(
        lifecycle: Optional[Callable] = None,
        registry: Optional[registries.GlobalRegistry] = None,
        standalone: bool = False,
        priority: int = 0,
        peering_name: Optional[str] = None,
        namespace: Optional[str] = None,
) -> Collection[asyncio.Task]:
    """
    Spawn all the tasks needed to run the operator.

    The tasks are properly inter-connected with the synchronisation primitives.
    """
    loop = asyncio.get_running_loop()

    # The freezer and the registry are scoped to this whole task-set, to sync them all.
    lifecycle = lifecycle if lifecycle is not None else lifecycles.get_default_lifecycle()
    registry = registry if registry is not None else registries.get_default_registry()
    event_queue = asyncio.Queue(loop=loop)
    freeze_flag = asyncio.Event(loop=loop)
    should_stop = asyncio.Event(loop=loop)
    tasks = []

    # A top-level task for external stopping by setting a stop-flag. Once set,
    # this task will exit, and thus all other top-level tasks will be cancelled.
    tasks.extend([
        loop.create_task(_stop_flag_checker(should_stop)),
    ])

    # K8s-event posting. Events are queued in-memory and posted in the background.
    # NB: currently, it is a global task, but can be made per-resource or per-object.
    tasks.extend([
        loop.create_task(posting.poster(
            event_queue=event_queue)),
    ])

    # Monitor the peers, unless explicitly disabled.
    ourselves: Optional[peering.Peer] = peering.Peer.detect(
        id=peering.detect_own_id(), priority=priority,
        standalone=standalone, namespace=namespace, name=peering_name,
    )
    if ourselves:
        tasks.extend([
            loop.create_task(peering.peers_keepalive(
                ourselves=ourselves)),
            loop.create_task(queueing.watcher(
                namespace=namespace,
                resource=ourselves.resource,
                handler=functools.partial(peering.peers_handler,
                                          ourselves=ourselves,
                                          freeze=freeze_flag))),  # freeze is set/cleared
        ])

    # Resource event handling, only once for every known resource (de-duplicated).
    for resource in registry.resources:
        tasks.extend([
            loop.create_task(queueing.watcher(
                namespace=namespace,
                resource=resource,
                handler=functools.partial(handling.custom_object_handler,
                                          lifecycle=lifecycle,
                                          registry=registry,
                                          resource=resource,
                                          event_queue=event_queue,
                                          freeze=freeze_flag))),  # freeze is only checked
        ])

    # On Ctrl+C or pod termination, cancel all tasks gracefully.
    if threading.current_thread() is threading.main_thread():
        loop.add_signal_handler(signal.SIGINT, should_stop.set)
        loop.add_signal_handler(signal.SIGTERM, should_stop.set)
    else:
        logger.warning("OS signals are ignored: running not in the main thread.")

    return tasks


async def run_tasks(root_tasks, *, ignored: Collection[asyncio.Task] = frozenset()):
    """
    Orchestrate the tasks and terminate them gracefully when needed.

    The root tasks are expected to run forever. Their number is limited. Once
    any of them exits, the whole operator and all other root tasks should exit.

    The root tasks, in turn, can spawn multiple sub-tasks of various purposes.
    They can be awaited, monitored, or fired-and-forgot.

    The hung tasks are those that were spawned during the operator runtime,
    and were not cancelled/exited on the root tasks termination. They are given
    some extra time to finish, after which they are forcely terminated too.

    .. note::
        Due to implementation details, every task created after the operator's
        startup is assumed to be a task or a sub-task of the operator.
        In the end, all tasks are forcely cancelled. Even if those tasks were
        created by other means. There is no way to trace who spawned what.
        Only the tasks that existed before the operator startup are ignored
        (for example, those that spawned the operator itself).
    """
    try:
        # Run the infinite tasks until one of them fails/exits (they never exit normally).
        root_done, root_pending = await _wait(root_tasks, return_when=asyncio.FIRST_COMPLETED)
    except asyncio.CancelledError:
        # If the operator is cancelled, propagate the cancellation to all the sub-tasks.
        # There is no graceful period: cancel as soon as possible, but allow them to finish.
        root_cancelled, root_left = await _stop(root_tasks, title="Root", cancelled=True)
        hung_tasks = await _all_tasks(ignored=ignored)
        hung_cancelled, hung_left = await _stop(hung_tasks, title="Hung", cancelled=True)
        raise
    else:
        # If the operator is intact, but one of the root tasks has exited (successfully or not),
        # cancel all the remaining root tasks, and gracefully exit other spawned sub-tasks.
        root_cancelled, root_left = await _stop(root_pending, title="Root", cancelled=False)
        hung_tasks = await _all_tasks(ignored=ignored)
        try:
            # After the root tasks are all gone, cancel any spawned sub-tasks (e.g. handlers).
            # TODO: assumption! the loop is not fully ours! find a way to cancel our spawned tasks.
            hung_done, hung_pending = await _wait(hung_tasks, timeout=5.0)
        except asyncio.CancelledError:
            # If the operator is cancelled, propagate the cancellation to all the sub-tasks.
            hung_cancelled, hung_left = await _stop(hung_tasks, title="Hung", cancelled=True)
            raise
        else:
            # If the operator is intact, but the timeout is reached, forcely cancel the sub-tasks.
            hung_cancelled, hung_left = await _stop(hung_pending, title="Hung", cancelled=False)

    # If succeeded or if cancellation is silenced, re-raise from failed tasks (if any).
    await _reraise(root_done | root_cancelled | hung_done | hung_cancelled)


async def _all_tasks(ignored: Collection[asyncio.Task] = frozenset()) -> Collection[asyncio.Task]:
    current_task = asyncio.current_task()
    return {task for task in asyncio.all_tasks()
            if task is not current_task and task not in ignored}


async def _wait(tasks, *, timeout=None, return_when=asyncio.ALL_COMPLETED):
    if not tasks:
        return set(), ()
    done, pending = await asyncio.wait(tasks, timeout=timeout, return_when=return_when)
    return done, pending


async def _stop(tasks, title, cancelled):
    if not tasks:
        logger.debug(f"{title} tasks stopping is skipped: no tasks given.")
        return set(), set()

    for task in tasks:
        task.cancel()

    # If the waiting (current) task is cancelled before the wait is over,
    # propagate the cancellation to all the awaited (sub-) tasks, and let them finish.
    try:
        done, pending = await asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
    except asyncio.CancelledError:
        # If the waiting (current) task is cancelled while propagating the cancellation
        # (i.e. double-cancelled), let it fail without graceful cleanup. It is urgent, it seems.
        pending = {task for task in tasks if not task.done()}
        are = 'are' if not pending else 'are not'
        why = 'double-cancelled at stopping' if cancelled else 'cancelled at stopping'
        logger.debug(f"{title} tasks {are} stopped: {why}; tasks left: {pending!r}")
        raise  # the repeated cancellation, handled specially.
    else:
        # If the cancellation is propagated normally and the awaited (sub-) tasks exited,
        # consider it as a successful cleanup.
        are = 'are' if not pending else 'are not'
        why = 'cancelled normally' if cancelled else 'finished normally'
        logger.debug(f"{title} tasks {are} stopped: {why}; tasks left: {pending!r}")
        return done, pending


async def _reraise(tasks):
    for task in tasks:
        try:
            task.result()  # can raise the regular (non-cancellation) exceptions.
        except asyncio.CancelledError:
            pass


async def _stop_flag_checker(should_stop):
    try:
        await should_stop.wait()
    except asyncio.CancelledError:
        pass  # operator is stopping for any other reason
    else:
        logger.debug("Stop-flag is raised. Operator is stopping.")


def create_tasks(loop: asyncio.AbstractEventLoop, *arg, **kwargs):
    """
    .. deprecated:: 1.0
        This is a synchronous interface to `spawn_tasks`.
        It is only kept for backward compatibility, as it was exposed
        via the public interface of the framework.
    """
    warnings.warn("kopf.create_tasks() is deprecated: "
                  "use kopf.spawn_tasks() or kopf.operator().",
                  DeprecationWarning)
    return loop.run_until_complete(spawn_tasks(*arg, **kwargs))



import kopf


def test_getting_default_lifecycle():
    lifecycle = kopf.get_default_lifecycle()
    assert lifecycle is kopf.lifecycles.asap


def test_setting_default_lifecycle():
    lifecycle_expected = lambda *args, **kwargs: None
    kopf.set_default_lifecycle(lifecycle_expected)
    lifecycle_actual = kopf.get_default_lifecycle()
    assert lifecycle_actual is lifecycle_expected


def test_resetting_default_lifecycle():
    lifecycle_distracting = lambda *args, **kwargs: None
    kopf.set_default_lifecycle(lifecycle_distracting)
    kopf.set_default_lifecycle(None)
    lifecycle_actual = kopf.get_default_lifecycle()
    assert lifecycle_actual is kopf.lifecycles.asap



"""
All the functions to manipulate the object finalization and deletion.

Finalizers are used to block the actual deletion until the finalizers
are removed, meaning that the operator has done all its duties
to "release" the object (e.g. cleanups; delete-handlers in our case).
"""

# A string marker to be put on the list of the finalizers to block
# the object from being deleted without the permission of the framework.
FINALIZER = 'kopf.zalando.org/KopfFinalizerMarker'
LEGACY_FINALIZER = 'KopfFinalizerMarker'


def is_deleted(body):
    return body.get('metadata', {}).get('deletionTimestamp', None) is not None


def has_finalizers(body):
    finalizers = body.get('metadata', {}).get('finalizers', [])
    return FINALIZER in finalizers or LEGACY_FINALIZER in finalizers


def append_finalizers(*, body, patch):
    if not has_finalizers(body=body):
        finalizers = body.get('metadata', {}).get('finalizers', [])
        patch.setdefault('metadata', {}).setdefault('finalizers', list(finalizers))
        patch['metadata']['finalizers'].append(FINALIZER)


def remove_finalizers(*, body, patch):
    if has_finalizers(body=body):
        finalizers = body.get('metadata', {}).get('finalizers', [])
        patch.setdefault('metadata', {}).setdefault('finalizers', list(finalizers))
        if LEGACY_FINALIZER in patch['metadata']['finalizers']:
            patch['metadata']['finalizers'].remove(LEGACY_FINALIZER)
        if FINALIZER in patch['metadata']['finalizers']:
            patch['metadata']['finalizers'].remove(FINALIZER)



"""
Only the tests from the watching (simulated) to the handling (substituted).

Excluded: the watching-streaming routines
(see ``tests_streaming.py`` and ``test_watching.py``).

Excluded: the causation and handling routines
(to be done later).

Used for internal control that the event queueing works are intended.
If the intentions change, the tests should be rewritten.
They are NOT part of the public interface of the framework.
"""
import asyncio
import weakref

import pytest

from kopf.reactor.queueing import watcher, EOS

# An overhead for the sync logic in async tests. Guesstimated empirically:
# 10ms is too fast, 200ms is too slow, 50-150ms is good enough (can vary).
CODE_OVERHEAD = 0.130


@pytest.mark.parametrize('uids, cnts, events', [

    pytest.param(['uid1'], [1], [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}}},
    ], id='single'),

    pytest.param(['uid1'], [3], [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'DELETED', 'object': {'metadata': {'uid': 'uid1'}}},
    ], id='multiple'),

    pytest.param(['uid1', 'uid2'], [3, 2], [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid2'}}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid2'}}},
        {'type': 'DELETED', 'object': {'metadata': {'uid': 'uid1'}}},
    ], id='mixed'),

])
@pytest.mark.usefixtures('watcher_limited')
async def test_watchevent_demultiplexing(worker_mock, timer, resource, handler,
                                         stream, events, uids, cnts):
    """ Verify that every unique uid goes into its own queue+worker, which are never shared. """

    # Inject the events of unique objects - to produce few streams/workers.
    stream.feed(events)

    # Run the watcher (near-instantly and test-blocking).
    with timer:
        await watcher(
            namespace=None,
            resource=resource,
            handler=handler,
        )

    # The streams are not cleared by the mocked worker, but the worker exits fast.
    assert timer.seconds < CODE_OVERHEAD

    # The handler must not be called by the watcher, only by the worker.
    # But the worker (even if mocked) must be called & awaited by the watcher.
    assert not handler.awaited
    assert not handler.called
    assert worker_mock.awaited

    # Are the worker-streams created by the watcher? Populated as expected?
    # One stream per unique uid? All events are sequential? EOS marker appended?
    assert worker_mock.call_count == len(uids)
    assert worker_mock.call_count == len(cnts)
    for uid, cnt, (args, kwargs) in zip(uids, cnts, worker_mock.call_args_list):
        key = kwargs['key']
        streams = kwargs['streams']
        assert kwargs['handler'] is handler
        assert key == (resource, uid)
        assert key in streams

        queue_events = []
        while not streams[key].watchevents.empty():
            queue_events.append(streams[key].watchevents.get_nowait())

        assert len(queue_events) == cnt + 1
        assert queue_events[-1] is EOS
        assert all(queue_event['object']['metadata']['uid'] == uid
                   for queue_event in queue_events[:-1])


@pytest.mark.parametrize('uids, vals, events', [

    pytest.param(['uid1'], ['b'], [
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'a'}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'b'}},
    ], id='the same'),

    pytest.param(['uid1', 'uid2'], ['a', 'b'], [
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'a'}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid2'}, 'spec': 'b'}},
    ], id='distinct'),

    pytest.param(['uid1', 'uid2', 'uid3'], ['e', 'd', 'f'], [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'a'}},
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid2'}, 'spec': 'b'}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'c'}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid2'}, 'spec': 'd'}},
        {'type': 'DELETED', 'object': {'metadata': {'uid': 'uid1'}, 'spec': 'e'}},
        {'type': 'DELETED', 'object': {'metadata': {'uid': 'uid3'}, 'spec': 'f'}},
    ], id='mixed'),

])
@pytest.mark.usefixtures('watcher_limited')
async def test_watchevent_batching(mocker, resource, handler, timer, stream, events, uids, vals):
    """ Verify that only the last event per uid is actually handled. """

    # Override the default timeouts to make the tests faster.
    mocker.patch('kopf.config.WorkersConfig.worker_idle_timeout', 0.5)
    mocker.patch('kopf.config.WorkersConfig.worker_batch_window', 0.1)
    mocker.patch('kopf.config.WorkersConfig.worker_exit_timeout', 0.5)

    # Inject the events of unique objects - to produce few streams/workers.
    stream.feed(events)

    # Run the watcher (near-instantly and test-blocking).
    with timer:
        await watcher(
            namespace=None,
            resource=resource,
            handler=handler,
        )

    # Significantly less than the queue getting timeout, but sufficient to run.
    # 2 <= 1 pull for the event chain + 1 pull for EOS. TODO: 1x must be enough.
    from kopf import config
    assert timer.seconds < config.WorkersConfig.worker_batch_window + CODE_OVERHEAD

    # Was the handler called at all? Awaited as needed for async fns?
    assert handler.awaited

    # Was it called only once per uid? Only with the latest event?
    assert handler.call_count == len(uids)
    assert handler.call_count == len(vals)
    for uid, val, (args, kwargs) in zip(uids, vals, handler.call_args_list):
        event = kwargs['event']
        assert event['object']['metadata']['uid'] == uid
        assert event['object']['spec'] == val


@pytest.mark.parametrize('unique, events', [

    pytest.param(1, [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'MODIFIED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'DELETED', 'object': {'metadata': {'uid': 'uid1'}}},
    ], id='the same'),

    pytest.param(2, [
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid1'}}},
        {'type': 'ADDED', 'object': {'metadata': {'uid': 'uid2'}}},
    ], id='distinct'),

])
@pytest.mark.usefixtures('watcher_in_background')
async def test_garbage_collection_of_streams(mocker, stream, events, unique, worker_spy):

    # Override the default timeouts to make the tests faster.
    mocker.patch('kopf.config.WorkersConfig.worker_idle_timeout', 0.5)
    mocker.patch('kopf.config.WorkersConfig.worker_batch_window', 0.1)
    mocker.patch('kopf.config.WorkersConfig.worker_exit_timeout', 0.5)
    mocker.patch('kopf.config.WatchersConfig.watcher_retry_delay', 1.0)  # to prevent src depletion

    # Inject the events of unique objects - to produce few streams/workers.
    stream.feed(events)

    # Give it a moment to populate the streams and spawn all the workers.
    # Intercept and remember _any_ seen dict of streams for further checks.
    while worker_spy.call_count < unique:
        await asyncio.sleep(0.001)  # give control to the loop
    streams = worker_spy.call_args_list[-1][1]['streams']

    # The mutable(!) streams dict is now populated with the objects' streams.
    assert len(streams) != 0  # usually 1, but can be 2+ if it is fast enough.

    # Weakly remember the stream's content to make sure it is gc'ed later.
    # Note: namedtuples are not referable due to __slots__/__weakref__ issues.
    refs = [weakref.ref(val) for wstream in streams.values() for val in wstream]
    assert all([ref() is not None for ref in refs])

    # Give the workers some time to finish waiting for the events.
    # Once the idle timeout, they will exit and gc their individual streams.
    from kopf import config
    await asyncio.sleep(config.WorkersConfig.worker_batch_window)  # depleting the queues.
    await asyncio.sleep(config.WorkersConfig.worker_idle_timeout)  # idling on empty queues.
    await asyncio.sleep(CODE_OVERHEAD)

    # The mutable(!) streams dict is now empty, i.e. garbage-collected.
    assert len(streams) == 0

    # Truly garbage-collected? Memory freed?
    assert all([ref() is None for ref in refs])


# TODO: also add tests for the depletion of the workers pools on cancellation (+timing)



"""
The reactor groups all modules to watch & process the low- & high-level events.

The low-level events are the kubernetes watch streams, received on every
object change, including the metainfo, status, etc.

The high-level events are the actually identified changes in the objects,
such as their creation, deletion, update both in general and per-field.
"""



import pytest

from kopf.structs.dicts import cherrypick


def test_overrides_existing_keys():
    src = {'ignored-key': 'src-val', 'tested-key': 'src-val'}
    dst = {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}
    cherrypick(src=src, dst=dst, fields=['tested-key'])
    assert dst == {'ignored-key': 'dst-val', 'tested-key': 'src-val'}


def test_adds_absent_dst_keys():
    src = {'ignored-key': 'src-val', 'tested-key': 'src-val'}
    dst = {'ignored-key': 'dst-val'}
    cherrypick(src=src, dst=dst, fields=['tested-key'])
    assert dst == {'ignored-key': 'dst-val', 'tested-key': 'src-val'}


def test_skips_absent_src_keys():
    src = {'ignored-key': 'src-val'}
    dst = {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}
    cherrypick(src=src, dst=dst, fields=['tested-key'])
    assert dst == {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}


def test_overrides_existing_subkeys():
    src = {'sub': {'ignored-key': 'src-val', 'tested-key': 'src-val'}}
    dst = {'sub': {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}}
    cherrypick(src=src, dst=dst, fields=['sub.tested-key'])
    assert dst == {'sub': {'ignored-key': 'dst-val', 'tested-key': 'src-val'}}


def test_adds_absent_dst_subkeys():
    src = {'sub': {'ignored-key': 'src-val', 'tested-key': 'src-val'}}
    dst = {'sub': {'ignored-key': 'dst-val'}}
    cherrypick(src=src, dst=dst, fields=['sub.tested-key'])
    assert dst == {'sub': {'ignored-key': 'dst-val', 'tested-key': 'src-val'}}


def test_skips_absent_src_subkeys():
    src = {'sub': {'ignored-key': 'src-val'}}
    dst = {'sub': {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}}
    cherrypick(src=src, dst=dst, fields=['sub.tested-key'])
    assert dst == {'sub': {'ignored-key': 'dst-val', 'tested-key': 'dst-val'}}


def test_ensures_dst_subdicts():
    src = {'sub': {'ignored-key': 'src-val', 'tested-key': 'src-val'}}
    dst = {}
    cherrypick(src=src, dst=dst, fields=['sub.tested-key'])
    assert dst == {'sub': {'tested-key': 'src-val'}}


def test_fails_on_nonmapping_src_key():
    src = {'sub': 'scalar-value'}
    dst = {'sub': {'ignored-key': 'src-val', 'tested-key': 'src-val'}}
    with pytest.raises(TypeError):
        cherrypick(src=src, dst=dst, fields=['sub.tested-key'])


def test_fails_on_nonmapping_dst_key():
    src = {'sub': {'ignored-key': 'src-val', 'tested-key': 'src-val'}}
    dst = {'sub': 'scalar-value'}
    with pytest.raises(TypeError):
        cherrypick(src=src, dst=dst, fields=['sub.tested-key'])



"""
Testing the handling of events on the top level.

As input:

* Mocked cause detection, with the cause artificially simulated for each test.
  The proper cause detection is tested elsewhere (see ``test_detection.py``).
* Registered handlers in a global registry. Each handler is a normal function,
  which calls a mock -- to ease the assertions.

As output, we check mocked calls on the following:

* ``asyncio.sleep()`` -- for delays.
* ``kopf.clients.patching.patch_obj()`` -- for patch content.
* ``kopf.clients.events.post_event()`` -- for events posted.
* Handler mocks -- whether they were or were not called with specific arguments.
* Captured logs.

The above inputs & outputs represent the expected user scenario
rather than the specific implementation of it.
Therefore, we do not mock/spy/intercept anything within the handling routines
(except for cause detection), leaving it as the implementation details.
Specifically, this internal chain of calls happens on every event:

* ``causation.detect_cause()`` -- tested separately in ``/tests/causation/``.
* ``handle_cause()``
* ``execute()``
* ``_execute()``
* ``_call_handler()``
* ``invocation.invoke()`` -- tested separately in ``/tests/invocations/``.

Some of these aspects are tested separately to be sure they indeed execute
all possible cases properly. In the top-level event handling, we assume they do,
and only check for the upper-level behaviour, not all of the input combinations.
"""
import copy
import dataclasses
from typing import Callable
from unittest.mock import Mock

import pytest

import kopf
from kopf.reactor.causation import Cause, RESUME


@dataclasses.dataclass(frozen=True, eq=False)
class K8sMocks:
    patch_obj: Mock
    post_event: Mock
    asyncio_sleep: Mock
    sleep_or_wait: Mock


@pytest.fixture(autouse=True)
def k8s_mocked(mocker, req_mock):
    # We mock on the level of our own K8s API wrappers, not the K8s client.
    return K8sMocks(
        patch_obj=mocker.patch('kopf.clients.patching.patch_obj'),
        post_event=mocker.patch('kopf.clients.events.post_event'),
        asyncio_sleep=mocker.patch('asyncio.sleep'),
        sleep_or_wait=mocker.patch('kopf.engines.sleeping.sleep_or_wait', return_value=None),
    )


@pytest.fixture(autouse=True)
def clear_default_registry():
    old_registry = kopf.get_default_registry()
    new_registry = kopf.GlobalRegistry()
    kopf.set_default_registry(new_registry)
    try:
        yield new_registry
    finally:
        kopf.set_default_registry(old_registry)


@pytest.fixture()
def registry(clear_default_registry):
    return clear_default_registry


@dataclasses.dataclass(frozen=True, eq=False, order=False)
class HandlersContainer:
    event_mock: Mock
    create_mock: Mock
    update_mock: Mock
    delete_mock: Mock
    resume_mock: Mock
    event_fn: Callable
    create_fn: Callable
    update_fn: Callable
    delete_fn: Callable
    resume_fn: Callable


@pytest.fixture()
def handlers(clear_default_registry):
    event_mock = Mock(return_value=None)
    create_mock = Mock(return_value=None)
    update_mock = Mock(return_value=None)
    delete_mock = Mock(return_value=None)
    resume_mock = Mock(return_value=None)

    @kopf.on.event('zalando.org', 'v1', 'kopfexamples', id='event_fn')
    async def event_fn(**kwargs):
        return event_mock(**kwargs)

    @kopf.on.create('zalando.org', 'v1', 'kopfexamples', id='create_fn', timeout=600)
    async def create_fn(**kwargs):
        return create_mock(**kwargs)

    @kopf.on.update('zalando.org', 'v1', 'kopfexamples', id='update_fn', timeout=600)
    async def update_fn(**kwargs):
        return update_mock(**kwargs)

    @kopf.on.delete('zalando.org', 'v1', 'kopfexamples', id='delete_fn', timeout=600)
    async def delete_fn(**kwargs):
        return delete_mock(**kwargs)

    @kopf.on.resume('zalando.org', 'v1', 'kopfexamples', id='resume_fn', timeout=600)
    async def resume_fn(**kwargs):
        return resume_mock(**kwargs)

    return HandlersContainer(
        event_mock=event_mock,
        create_mock=create_mock,
        update_mock=update_mock,
        delete_mock=delete_mock,
        resume_mock=resume_mock,
        event_fn=event_fn,
        create_fn=create_fn,
        update_fn=update_fn,
        delete_fn=delete_fn,
        resume_fn=resume_fn,
    )


@pytest.fixture()
def extrahandlers(clear_default_registry, handlers):
    event_mock = Mock(return_value=None)
    create_mock = Mock(return_value=None)
    update_mock = Mock(return_value=None)
    delete_mock = Mock(return_value=None)
    resume_mock = Mock(return_value=None)

    @kopf.on.event('zalando.org', 'v1', 'kopfexamples', id='event_fn2')
    async def event_fn2(**kwargs):
        return event_mock(**kwargs)

    @kopf.on.create('zalando.org', 'v1', 'kopfexamples', id='create_fn2')
    async def create_fn2(**kwargs):
        return create_mock(**kwargs)

    @kopf.on.update('zalando.org', 'v1', 'kopfexamples', id='update_fn2')
    async def update_fn2(**kwargs):
        return update_mock(**kwargs)

    @kopf.on.delete('zalando.org', 'v1', 'kopfexamples', id='delete_fn2')
    async def delete_fn2(**kwargs):
        return delete_mock(**kwargs)

    @kopf.on.resume('zalando.org', 'v1', 'kopfexamples', id='resume_fn2')
    async def resume_fn2(**kwargs):
        return resume_mock(**kwargs)

    return HandlersContainer(
        event_mock=event_mock,
        create_mock=create_mock,
        update_mock=update_mock,
        delete_mock=delete_mock,
        resume_mock=resume_mock,
        event_fn=event_fn2,
        create_fn=create_fn2,
        update_fn=update_fn2,
        delete_fn=delete_fn2,
        resume_fn=resume_fn2,
    )


@pytest.fixture()
def cause_mock(mocker, resource):

    # Use everything from a mock, but use the passed `patch` dict as is.
    # The event handler passes its own accumulator, and checks/applies it later.
    def new_detect_fn(**kwargs):

        # Avoid collision of our mocked values with the passed kwargs.
        original_event = kwargs.pop('event', None)
        original_body = kwargs.pop('body', None)
        original_diff = kwargs.pop('diff', None)
        original_new = kwargs.pop('new', None)
        original_old = kwargs.pop('old', None)
        original_digest = kwargs.pop('digest', None)
        event = mock.event if mock.event is not None else original_event
        initial = bool(event == RESUME)
        body = copy.deepcopy(mock.body) if mock.body is not None else original_body
        diff = copy.deepcopy(mock.diff) if mock.diff is not None else original_diff
        new = copy.deepcopy(mock.new) if mock.new is not None else original_new
        old = copy.deepcopy(mock.old) if mock.old is not None else original_old
        digest = copy.deepcopy(mock.digest) if mock.digest is not None else original_digest

        # Remove requires_finalizer from kwargs as it shouldn't be passed to the Cause object
        kwargs.pop('requires_finalizer', None)

        # Pass through kwargs: resource, logger, patch, diff, old, new.
        # I.e. everything except what we mock: event & body.
        cause = Cause(
            event=event,
            initial=initial,
            body=body,
            diff=diff,
            new=new,
            old=old,
            digest=digest,
            **kwargs)

        # Needed for the k8s-event creation, as they are attached to objects.
        body.setdefault('apiVersion', f'{resource.group}/{resource.version}')
        body.setdefault('kind', 'KopfExample')  # TODO: resource.???
        body.setdefault('metadata', {}).setdefault('namespace', 'some-namespace')
        body.setdefault('metadata', {}).setdefault('name', 'some-name')
        body.setdefault('metadata', {}).setdefault('uid', 'some-uid')

        return cause

    # Substitute the real cause detector with out own mock-based one.
    mocker.patch('kopf.reactor.causation.detect_cause', new=new_detect_fn)

    # The mock object stores some values later used by the factory substitute.
    mock = mocker.Mock(spec_set=['event', 'body', 'diff', 'new', 'old', 'digest'])
    mock.event = None
    mock.body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    mock.diff = None
    mock.new = None
    mock.old = None
    mock.digest = None
    return mock



import asyncio
import datetime
import logging

import pykube
import requests

from kopf import config
from kopf.clients import auth
from kopf.structs import hierarchies

logger = logging.getLogger(__name__)

MAX_MESSAGE_LENGTH = 1024
CUT_MESSAGE_INFIX = '...'


async def post_event(*, obj=None, ref=None, type, reason, message=''):
    """
    Issue an event for the object.

    This is where they can also be accumulated, aggregated, grouped,
    and where the rate-limits should be maintained. It can (and should)
    be done by the client library, as it is done in the Go client.
    """

    # Object reference - similar to the owner reference, but different.
    if obj is not None and ref is not None:
        raise TypeError("Only one of obj= and ref= is allowed for a posted event. Got both.")
    if obj is None and ref is None:
        raise TypeError("One of obj= and ref= is required for a posted event. Got none.")
    if ref is None:
        ref = hierarchies.build_object_reference(obj)

    now = datetime.datetime.utcnow()

    # See #164. For cluster-scoped objects, use the current namespace from the current context.
    # It could be "default", but in some systems, we are limited to one specific namespace only.
    namespace = ref.get('namespace') or auth.get_pykube_cfg().namespace
    if not ref.get('namespace'):
        ref = dict(ref, namespace=namespace)

    # Prevent a common case of event posting errors but shortening the message.
    if len(message) > MAX_MESSAGE_LENGTH:
        infix = CUT_MESSAGE_INFIX
        prefix = message[:MAX_MESSAGE_LENGTH // 2 - (len(infix) // 2)]
        suffix = message[-MAX_MESSAGE_LENGTH // 2 + (len(infix) - len(infix) // 2):]
        message = f'{prefix}{infix}{suffix}'

    body = {
        'metadata': {
            'namespace': namespace,
            'generateName': 'kopf-event-',
        },

        'action': 'Action?',
        'type': type,
        'reason': reason,
        'message': message,

        'reportingComponent': 'kopf',
        'reportingInstance': 'dev',
        'source' : {'component': 'kopf'},  # used in the "From" column in `kubectl describe`.

        'involvedObject': ref,

        'firstTimestamp': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' -- seen in `kubectl describe ...`
        'lastTimestamp': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z' - seen in `kubectl get events`
        'eventTime': now.isoformat() + 'Z',  # '2019-01-28T18:25:03.000000Z'
    }

    try:
        api = auth.get_pykube_api()
        obj = pykube.Event(api, body)

        loop = asyncio.get_running_loop()
        await loop.run_in_executor(config.WorkersConfig.get_syn_executor(), obj.create)

    except (requests.exceptions.HTTPError, pykube.exceptions.HTTPError) as e:
        # Events are helpful but auxiliary, they should not fail the handling cycle.
        # Yet we want to notice that something went wrong (in logs).
        logger.warning("Failed to post an event. Ignoring and continuing. "
                       f"Error: {e!r}. "
                       f"Event: type={type!r}, reason={reason!r}, message={message!r}.")



import os.path
import subprocess
import time

import pytest

import kopf.testing

crd_yaml = os.path.relpath(os.path.join(os.path.dirname(__file__), '..', 'crd.yaml'))
obj_yaml = os.path.relpath(os.path.join(os.path.dirname(__file__), '..', 'obj.yaml'))
example_py = os.path.relpath(os.path.join(os.path.dirname(__file__), 'example.py'))


@pytest.fixture(autouse=True)
def crd_exists():
    subprocess.run(f"kubectl apply -f {crd_yaml}",
                   check=True, timeout=10, capture_output=True, shell=True)


@pytest.fixture(autouse=True)
def obj_absent():
    # Operator is not running in fixtures, so we need a force-delete (or this patch).
    subprocess.run(['kubectl', 'patch', '-f', obj_yaml,
                    '-p', '{"metadata":{"finalizers":[]}}',
                    '--type', 'merge'],
                   check=False, timeout=10, capture_output=True)
    subprocess.run(f"kubectl delete -f {obj_yaml}",
                   check=False, timeout=10, capture_output=True, shell=True)


def test_resource_lifecycle(mocker):

    # To prevent lengthy threads in the loop executor when the process exits.
    mocker.patch('kopf.config.WatchersConfig.default_stream_timeout', 10)

    # Run an operator and simulate some activity with the operated resource.
    with kopf.testing.KopfRunner(['run', '--verbose', '--standalone', example_py], timeout=60) as runner:
        subprocess.run(f"kubectl create -f {obj_yaml}", shell=True, check=True)
        time.sleep(5)  # give it some time to react
        subprocess.run(f"kubectl delete -f {obj_yaml}", shell=True, check=True)
        time.sleep(1)  # give it some time to react

    # Ensure that the operator did not die on start, or during the operation.
    assert runner.exception is None
    assert runner.exit_code == 0

    # There are usually more than these messages, but we only check for the certain ones.
    assert '[default/kopf-example-1] Creation event:' in runner.stdout
    assert '[default/kopf-example-1] Something was logged here.' in runner.stdout



"""
Just to make sure that asyncio tests are configured properly.
"""
import asyncio

_async_was_executed = False


async def test_async_tests_are_enabled(timer):
    global _async_was_executed
    _async_was_executed = True  # asserted in a sync-test below.

    with timer as t:
        await asyncio.sleep(0.5)

    assert t.seconds > 0.5  # real sleep


async def test_async_mocks_are_enabled(timer, mocker):
    p = mocker.patch('asyncio.sleep')
    with timer as t:
        await asyncio.sleep(1.0)

    assert p.called
    assert p.awaited
    assert t.seconds < 0.01  # mocked sleep


def test_async_test_was_executed_and_awaited():
    assert _async_was_executed



import pytest

from kopf.structs.finalizers import FINALIZER, LEGACY_FINALIZER
from kopf.structs.finalizers import append_finalizers, remove_finalizers
from kopf.structs.finalizers import is_deleted, has_finalizers


def test_finalizer_is_fqdn():
    assert FINALIZER.startswith('kopf.zalando.org/')


@pytest.mark.parametrize('expected, body', [
    pytest.param(True, {'metadata': {'deletionTimestamp': '2020-12-31T23:59:59'}}, id='time'),
    pytest.param(False, {'metadata': {'deletionTimestamp': None}}, id='none'),
    pytest.param(False, {'metadata': {}}, id='no-field'),
    pytest.param(False, {}, id='no-metadata'),
])
def test_is_deleted(expected, body):
    result = is_deleted(body=body)
    assert result == expected


@pytest.mark.parametrize('expected, body', [
    pytest.param(False, {}, id='no-metadata'),
    pytest.param(False, {'metadata': {}}, id='no-finalizers'),
    pytest.param(False, {'metadata': {'finalizers': []}}, id='empty'),
    pytest.param(False, {'metadata': {'finalizers': ['other']}}, id='others'),
    pytest.param(True, {'metadata': {'finalizers': [FINALIZER]}}, id='normal'),
    pytest.param(True, {'metadata': {'finalizers': [LEGACY_FINALIZER]}}, id='legacy'),
    pytest.param(True, {'metadata': {'finalizers': ['other', FINALIZER]}}, id='mixed'),
])
def test_has_finalizers(expected, body):
    result = has_finalizers(body=body)
    assert result == expected


def test_append_finalizers_to_others():
    body = {'metadata': {'finalizers': ['other1', 'other2']}}
    patch = {}
    append_finalizers(body=body, patch=patch)
    assert patch == {'metadata': {'finalizers': ['other1', 'other2', FINALIZER]}}


def test_append_finalizers_to_empty():
    body = {}
    patch = {}
    append_finalizers(body=body, patch=patch)
    assert patch == {'metadata': {'finalizers': [FINALIZER]}}


def test_append_finalizers_when_present():
    body = {'metadata': {'finalizers': ['other1', FINALIZER, 'other2']}}
    patch = {}
    append_finalizers(body=body, patch=patch)
    assert patch == {}


@pytest.mark.parametrize('finalizer', [
    pytest.param(LEGACY_FINALIZER, id='legacy'),
    pytest.param(FINALIZER, id='normal'),
])
def test_remove_finalizers_keeps_others(finalizer):
    body = {'metadata': {'finalizers': ['other1', finalizer, 'other2']}}
    patch = {}
    remove_finalizers(body=body, patch=patch)
    assert patch == {'metadata': {'finalizers': ['other1', 'other2']}}


def test_remove_finalizers_when_absent():
    body = {'metadata': {'finalizers': ['other1', 'other2']}}
    patch = {}
    remove_finalizers(body=body, patch=patch)
    assert patch == {}


def test_remove_finalizers_when_empty():
    body = {}
    patch = {}
    remove_finalizers(body=body, patch=patch)
    assert patch == {}



import functools

import pytest

from kopf import SimpleRegistry
from kopf.reactor.registries import get_callable_id


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn():
    pass


@pytest.fixture(params=[
    'some-field.sub-field',
    ['some-field', 'sub-field'],
    ('some-field', 'sub-field'),
], ids=['str', 'list', 'tuple'])
def field(request):
    return request.param


def test_id_of_simple_function():
    fn_id = get_callable_id(some_fn)
    assert fn_id == 'some_fn'


def test_id_of_single_partial():
    partial_fn = functools.partial(some_fn)

    fn_id = get_callable_id(partial_fn)
    assert fn_id == 'some_fn'


def test_id_of_double_partial():
    partial1_fn = functools.partial(some_fn)
    partial2_fn = functools.partial(partial1_fn)

    fn_id = get_callable_id(partial2_fn)
    assert fn_id == 'some_fn'


def test_id_of_single_wrapper():

    @functools.wraps(some_fn)
    def wrapped_fn():
        pass

    fn_id = get_callable_id(wrapped_fn)
    assert fn_id == 'some_fn'


def test_id_of_double_wrapper():

    @functools.wraps(some_fn)
    def wrapped1_fn():
        pass

    @functools.wraps(wrapped1_fn)
    def wrapped2_fn():
        pass

    fn_id = get_callable_id(wrapped2_fn)
    assert fn_id == 'some_fn'


def test_id_of_lambda():
    some_lambda = lambda: None

    fn_id = get_callable_id(some_lambda)
    assert fn_id.startswith(f'lambda:{__file__}:')


def test_with_no_hints(mocker):
    get_fn_id = mocker.patch('kopf.reactor.registries.get_callable_id', return_value='some-id')

    registry = SimpleRegistry()
    registry.register(some_fn)
    handlers = registry.get_cause_handlers(mocker.MagicMock())

    assert get_fn_id.called

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn
    assert handlers[0].id == 'some-id'


def test_with_prefix(mocker):
    get_fn_id = mocker.patch('kopf.reactor.registries.get_callable_id', return_value='some-id')

    registry = SimpleRegistry(prefix='some-prefix')
    registry.register(some_fn)
    handlers = registry.get_cause_handlers(mocker.MagicMock())

    assert get_fn_id.called

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn
    assert handlers[0].id == 'some-prefix/some-id'


def test_with_suffix(mocker, field):
    get_fn_id = mocker.patch('kopf.reactor.registries.get_callable_id', return_value='some-id')
    diff = [('add', ('some-field', 'sub-field'), 'old', 'new')]

    registry = SimpleRegistry()
    registry.register(some_fn, field=field)
    handlers = registry.get_cause_handlers(mocker.MagicMock(diff=diff))

    assert get_fn_id.called

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn
    assert handlers[0].id == 'some-id/some-field.sub-field'


def test_with_prefix_and_suffix(mocker, field):
    get_fn_id = mocker.patch('kopf.reactor.registries.get_callable_id', return_value='some-id')
    diff = [('add', ('some-field', 'sub-field'), 'old', 'new')]

    registry = SimpleRegistry(prefix='some-prefix')
    registry.register(some_fn, field=field)
    handlers = registry.get_cause_handlers(mocker.MagicMock(diff=diff))

    assert get_fn_id.called

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn
    assert handlers[0].id == 'some-prefix/some-id/some-field.sub-field'


def test_with_explicit_id_and_prefix_and_suffix(mocker, field):
    get_fn_id = mocker.patch('kopf.reactor.registries.get_callable_id', return_value='some-id')
    diff = [('add', ('some-field', 'sub-field'), 'old', 'new')]

    registry = SimpleRegistry(prefix='some-prefix')
    registry.register(some_fn, id='explicit-id', field=field)
    handlers = registry.get_cause_handlers(mocker.MagicMock(diff=diff))

    assert not get_fn_id.called

    assert len(handlers) == 1
    assert handlers[0].fn is some_fn
    assert handlers[0].id == 'some-prefix/explicit-id/some-field.sub-field'



import asyncio
import concurrent.futures
import logging
from typing import Optional

from kopf.engines import logging as logging_engine

format = '[%(asctime)s] %(name)-20.20s [%(levelname)-8.8s] %(message)s'


LOGLEVEL_INFO = 20
""" Event loglevel to log all events. """
LOGLEVEL_WARNING = 30
""" Event loglevel to log all events except informational. """
LOGLEVEL_ERROR = 40
""" Event loglevel to log only errors and critical events. """
LOGLEVEL_CRITICAL = 50
""" Event loglevel to log only critical events(basically - no events). """


def configure(debug=None, verbose=None, quiet=None):
    log_level = 'DEBUG' if debug or verbose else 'WARNING' if quiet else 'INFO'

    logger = logging.getLogger()
    handler = logging.StreamHandler()
    formatter = logging_engine.ObjectPrefixingFormatter(format)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(log_level)

    # Configure the Kubernetes client defaults according to our settings.
    try:
        import kubernetes
    except ImportError:
        pass
    else:
        config = kubernetes.client.configuration.Configuration()
        config.logger_format = format
        config.logger_file = None  # once again after the constructor to re-apply the formatter
        config.debug = debug
        kubernetes.client.configuration.Configuration.set_default(config)

    # Kubernetes client is as buggy as hell: it adds its own stream handlers even in non-debug mode,
    # does not respect the formatting, and dumps too much of the low-level info.
    if not debug:
        logger = logging.getLogger("urllib3")
        del logger.handlers[1:]  # everything except the default NullHandler

    # Prevent the low-level logging unless in the debug verbosity mode. Keep only the operator's messages.
    logging.getLogger('urllib3').propagate = debug
    logging.getLogger('asyncio').propagate = debug
    logging.getLogger('kubernetes').propagate = debug

    loop = asyncio.get_event_loop()
    loop.set_debug(debug)


class EventsConfig:
    """
    Used to configure events sending behaviour.
    """

    events_loglevel = LOGLEVEL_INFO
    """ What events should be logged. """


class WorkersConfig:
    """
    Used as single point of configuration for kopf.reactor.
    """

    threadpool_executor: Optional[concurrent.futures.ThreadPoolExecutor] = None

    queue_workers_limit: Optional[int] = None  # if None, there is no limits to workers number
    """ How many workers can be running simultaneously on per-object event queue. """

    synchronous_tasks_threadpool_limit: Optional[int] = None  # if None, calculated by ThreadPoolExecutor based on cpu count
    """ How many threads in total can be running simultaneously to handle any non-async tasks."""

    worker_idle_timeout: float = 5.0
    """ How long does a worker can idle before exiting and garbage-collecting."""

    worker_batch_window: float = 0.1
    """ How fast/slow does a worker deplete the queue when an event is received."""

    worker_exit_timeout: float = 2.0
    """ How long does a worker can work on watcher exit before being cancelled. """

    @staticmethod
    def get_syn_executor() -> concurrent.futures.ThreadPoolExecutor:
        if not WorkersConfig.threadpool_executor:
            WorkersConfig.threadpool_executor = concurrent.futures.ThreadPoolExecutor(
                max_workers=WorkersConfig.synchronous_tasks_threadpool_limit
            )
        return WorkersConfig.threadpool_executor

    @staticmethod
    def set_synchronous_tasks_threadpool_limit(new_limit: int):
        """
        Call this static method at any time to change synchronous_tasks_threadpool_limit in runtime.
        """
        if new_limit < 1:
            raise ValueError('Can`t set threadpool limit lower than 1')

        WorkersConfig.synchronous_tasks_threadpool_limit = new_limit
        if WorkersConfig.threadpool_executor:
            WorkersConfig.threadpool_executor._max_workers = new_limit


class WatchersConfig:
    """
    Used to configure the K8s API watchers and streams.
    """

    default_stream_timeout = None
    """ The maximum duration of one streaming request. Patched in some tests. """

    watcher_retry_delay = 0.1
    """ How long should a pause be between watch requests (to prevent flooding). """



import asyncio
import functools

import click

from kopf import config
from kopf.clients import auth
from kopf.engines import peering
from kopf.reactor import running
from kopf.utilities import loaders


def cli_login():
    try:
        auth.login(verify=True)
    except auth.LoginError as e:
        raise click.ClickException(str(e))
    except auth.AccessError as e:
        raise click.ClickException(str(e))


def logging_options(fn):
    """ A decorator to configure logging in all command in the same way."""
    @click.option('-v', '--verbose', is_flag=True)
    @click.option('-d', '--debug', is_flag=True)
    @click.option('-q', '--quiet', is_flag=True)
    @functools.wraps(fn)  # to preserve other opts/args
    def wrapper(verbose, quiet, debug, *args, **kwargs):
        config.configure(debug=debug, verbose=verbose, quiet=quiet)
        return fn(*args, **kwargs)

    return wrapper


@click.version_option(prog_name='kopf')
@click.group(name='kopf', context_settings=dict(
    auto_envvar_prefix='KOPF',
))
def main():
    pass


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('--standalone', is_flag=True, default=False)
@click.option('--dev', 'priority', type=int, is_flag=True, flag_value=666)
@click.option('-P', '--peering', 'peering_name', type=str, default=None, envvar='KOPF_RUN_PEERING')
@click.option('-p', '--priority', type=int, default=0)
@click.option('-m', '--module', 'modules', multiple=True)
@click.argument('paths', nargs=-1)
def run(paths, modules, peering_name, priority, standalone, namespace):
    """ Start an operator process and handle all the requests. """
    cli_login()
    loaders.preload(
        paths=paths,
        modules=modules,
    )
    return running.run(
        standalone=standalone,
        namespace=namespace,
        priority=priority,
        peering_name=peering_name,
    )


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('-i', '--id', type=str, default=None)
@click.option('--dev', 'priority', flag_value=666)
@click.option('-P', '--peering', 'peering_name', type=str, default=None, envvar='KOPF_FREEZE_PEERING')
@click.option('-p', '--priority', type=int, default=100)
@click.option('-t', '--lifetime', type=int, required=True)
@click.option('-m', '--message', type=str)
def freeze(id, message, lifetime, namespace, peering_name, priority):
    """ Freeze the resource handling in the cluster. """
    cli_login()
    ourserlves = peering.Peer(
        id=id or peering.detect_own_id(),
        name=peering_name,
        namespace=namespace,
        priority=priority,
        lifetime=lifetime,
    )
    loop = asyncio.get_event_loop()
    loop.run_until_complete(ourserlves.keepalive())


@main.command()
@logging_options
@click.option('-n', '--namespace', default=None)
@click.option('-i', '--id', type=str, default=None)
@click.option('-P', '--peering', 'peering_name', type=str, default=None, envvar='KOPF_RESUME_PEERING')
def resume(id, namespace, peering_name):
    """ Resume the resource handling in the cluster. """
    cli_login()
    ourselves = peering.Peer(
        id=id or peering.detect_own_id(),
        name=peering_name,
        namespace=namespace,
    )
    loop = asyncio.get_event_loop()
    loop.run_until_complete(ourselves.disappear())
